import numpy as np
import pandas as pd
import xarray as xr
from copy import copy, deepcopy
from scipy.interpolate import interpn

from foxes.core import States, get_engine
from foxes.utils import import_module
from foxes.data import STATES, StaticData
from foxes.utils.wind_dir import uv2wd, wd2uv
from foxes.config import config, get_input_path
import foxes.variables as FV
import foxes.constants as FC


def _read_nc_file(
    fpath,
    coords,
    vars,
    nc_engine,
    sel,
    isel,
    minimal,
    drop_vars=None,
    check_input_nans=True,
):
    """Helper function for nc file reading"""
    with xr.open_dataset(fpath, drop_variables=drop_vars, engine=nc_engine) as data:
        for c in coords:
            if c is not None and c not in data.sizes:
                raise KeyError(
                    f"Missing coordinate '{c}' in file {fpath}, got: {list(data.sizes.keys())}"
                )

        if minimal:
            data = data[coords[0]].to_numpy()
        else:
            data = data[vars]
            data.attrs = {}
            if isel is not None and len(isel):
                data = data.isel(**isel)
            if sel is not None and len(sel):
                data = data.sel(**sel)
            assert min(data.sizes.values()) > 0, (
                f"States: No data in file {fpath}, isel={isel}, sel={sel}, resulting sizes={data.sizes}"
            )
            if check_input_nans:
                for v, d in data.data_vars.items():
                    sel = np.isnan(d.to_numpy())
                    if sel.any():
                        i = tuple([j[0] for j in np.where(sel)])
                        print("\n\nError: NaN data found in input data:")
                        print(f"  File: {fpath}\n")
                        print(f"  Variable: {v}")
                        for ic, c in enumerate(d.dims):
                            print(f"  {c}: {data[c].to_numpy()[i[ic]]}")
                        print("\n\n")
                        raise ValueError(
                            f"States: NaN data found in input data for variable '{v}' with dims {d.dims} in file {fpath} at index {i}"
                        )

    return data


class DatasetStates(States):
    """
    Abstract base class for heterogeneous ambient states that
    are based on data from NetCDF files or an xarray Dataset.

    Attributes
    ----------
    data_source: str or xarray.Dataset
        The data or the file search pattern, should end with
        suffix '.nc'. One or many files.
    ovars: list of str
        The output variables
    var2ncvar: dict
        Mapping from variable names to variable names
        in the nc file
    fixed_vars: dict
        Uniform values for output variables, instead
        of reading from data
    load_mode: str
        The load mode, choices: preload, lazy, fly.
        preload loads all data during initialization,
        lazy lazy-loads the data using dask, and fly
        reads only states index and weights during initialization
        and then opens the relevant files again within
        the chunk calculation
    time_format: str
        The datetime parsing format string
    sel: dict, optional
        Subset selection via xr.Dataset.sel()
    isel: dict, optional
        Subset selection via xr.Dataset.isel()
    weight_factor: float
        The factor to multiply the weights with
    check_times: bool
        Whether to check the time coordinates for consistency
    check_input_nans: bool
        Whether to check input data for NaNs

    :group: input.states

    """

    def __init__(
        self,
        data_source,
        output_vars,
        var2ncvar={},
        fixed_vars={},
        load_mode="preload",
        time_format=r"%Y-%m-%d_%H:%M:%S",
        sel=None,
        isel=None,
        weight_factor=None,
        check_times=True,
        check_input_nans=True,
        **kwargs,
    ):
        """
        Constructor.

        Parameters
        ----------
        data_source: str or xarray.Dataset
            The data or the file search pattern, should end with
            suffix '.nc'. One or many files.
        output_vars: list of str
            The output variables
        var2ncvar: dict, optional
            Mapping from variable names to variable names
            in the nc file
        fixed_vars: dict, optional
            Uniform values for output variables, instead
            of reading from data
        load_mode: str
            The load mode, choices: preload, lazy, fly.
            preload loads all data during initialization,
            lazy lazy-loads the data using dask, and fly
            reads only states index and weights during initialization
            and then opens the relevant files again within
            the chunk calculation
        time_format: str
            The datetime parsing format string
        sel: dict, optional
            Subset selection via xr.Dataset.sel()
        isel: dict, optional
            Subset selection via xr.Dataset.isel()
        weight_factor: float, optional
            The factor to multiply the weights with
        check_times: bool
            Whether to check the time coordinates for consistency
        check_input_nans: bool
            Whether to check input data for NaNs
        kwargs: dict, optional
            Additional arguments for the base class

        """
        super().__init__(**kwargs)

        self.ovars = list(output_vars)
        self.fixed_vars = fixed_vars
        self.load_mode = load_mode
        self.var2ncvar = var2ncvar
        self.time_format = time_format
        self.sel = sel
        self.isel = isel
        self.weight_factor = weight_factor
        self.check_times = check_times
        self.check_input_nans = check_input_nans

        self._N = None
        self._inds = None
        self.__data_source = data_source

    @property
    def data_source(self):
        """
        The data source

        Returns
        -------
        s: object
            The data source

        """
        if self.load_mode in ["preload", "fly"] and self.running:
            raise ValueError(
                f"States '{self.name}': Cannot access data_source while running for load mode '{self.load_mode}'"
            )
        return self.__data_source

    def _read_ds(self, ds, cmap, variables, verbosity=0):
        """
        Helper function for _get_data, extracts data from the original Dataset.

        Parameters
        ----------
        ds: xarray.Dataset
            The Dataset to read data from
        cmap: dict
            A mapping from foxes variable names to Dataset dimension names
        variables: list of str
            The variables to extract from the Dataset
        verbosity: int
            The verbosity level, 0 = silent

        Returns
        -------
        coords: dict
            keys: Foxes variable names, values: 1D coordinate value arrays
        data: dict
            The extracted data, keys are variable names,
            values are tuples (dims, data_array)
            where dims is a tuple of dimension names and
            data_array is a numpy.ndarray with the data values

        """
        data = {}
        for v in variables:
            w = self.var2ncvar.get(v, v)
            if w in ds.data_vars:
                d = ds[w]
                i = [d.dims.index(c) for c in cmap.values() if c in d.dims]
                assert len(i) == len(d.dims), (
                    f"States '{self.name}': Variable '{w}' has dimensions {d.dims}, but not all of them are in the coordinate mapping {cmap}"
                )
                dms = tuple([v for v, c in cmap.items() if c in d.dims])
                j = list(range(len(i)))
                if i == j:
                    data[v] = (dms, d.to_numpy())
                elif len(i) == 2:
                    data[v] = (dms, np.swapaxes(d.to_numpy(), 0, 1))
                else:
                    data[v] = (dms, np.moveaxis(d.to_numpy(), i, j))
            else:
                raise KeyError(
                    f"States '{self.name}': Variable '{w}' not found in data source '{self.data_source}', available variables: {list(ds.data_vars)}"
                )

        coords = {v: ds[c].to_numpy() for v, c in cmap.items() if c in ds.coords}

        if verbosity > 1:
            if len(coords):
                print(f"\n{self.name}: Coordinate ranges")
                for c, d in coords.items():
                    print(f"  {c}: {np.min(d)} --> {np.max(d)}")
            print(f"\n{self.name}: Data ranges")
            for v, d in data.items():
                nn = np.sum(np.isnan(d))
                print(
                    f"  {v}: {np.nanmin(d)} --> {np.nanmax(d)}, nans: {nn} ({100 * nn / len(d.flat):.2f}%)"
                )

        return coords, data

    def _get_data(self, ds, cmap, variables, verbosity=0):
        """
        Gets the data from the Dataset and prepares it for calculations.

        Parameters
        ----------
        ds: xarray.Dataset
            The Dataset to read data from
        cmap: dict
            A mapping from foxes variable names to Dataset dimension names
        variables: list of str
            The variables to extract from the Dataset
        verbosity: int
            The verbosity level, 0 = silent

        Returns
        -------
        coords: dict
            keys: Foxes variable names, values: 1D coordinate value arrays
        data: dict
            The extracted data, keys are dimension tuples,
            values are tuples (DATA key, variables, data_array)
            where DATA key is the name in the mdata object,
            variables is a list of variable names, and
            data_array is a numpy.ndarray with the data values,
            the last dimension corresponds to the variables
        weights: numpy.ndarray or None
            The weights array, if only state dependent, otherwise
            weights are among data. Shape: (n_states,)

        """
        coords, data0 = self._read_ds(ds, cmap, variables, verbosity=verbosity)

        weights = None
        if FV.WEIGHT in variables:
            assert FV.WEIGHT in data0, (
                f"States '{self.name}': Missing weights variable '{FV.WEIGHT}' in data, found {sorted(list(data0.keys()))}"
            )
            if self.weight_factor is not None:
                data0[FV.WEIGHT][1] *= self.weight_factor
            if data0[FV.WEIGHT][0] == (FC.STATE,):
                weights = data0.pop(FV.WEIGHT)[1]

        data = {}  # dim: [DATA key, variables, data array]
        for v, (dims, d) in data0.items():
            if dims not in data:
                i = len(data)
                data[dims] = [self.var(f"data{i}"), [], []]
            data[dims][1].append(v)
            data[dims][2].append(d)
        for dims in data.keys():
            data[dims][2] = np.stack(data[dims][2], axis=-1)
        data = {
            tuple(list(dims) + [f"vars{i}"]): d
            for i, (dims, d) in enumerate(data.items())
        }
        return coords, data, weights

    def preproc_first(
        self, algo, data, cmap, vars, bounds_extra_space, height_bounds, verbosity=0
    ):
        """
        Preprocesses the first file.

        Parameters
        ----------
        algo: foxes.core.Algorithm
            The calculation algorithm
        data: xarray.Dataset
            The dataset to preprocess
        cmap: dict
            A mapping from foxes variable names to Dataset dimension names
        vars: list
            The list of variable names
        bounds_extra_space: float or str, optional
            The extra space, either float in m,
            or str for units of D, e.g. '2.5D'
        height_bounds: tuple, optional
            The (h_min, h_max) height bounds in m. Defaults to H +/-
        verbosity: int
            The verbosity level, 0 = silent

        """

        # find vertical bounds:
        if FV.H in cmap:
            if height_bounds is None:
                H = algo.farm.get_hub_heights(algo)
                D = algo.farm.get_rotor_diameters(algo)
                H = np.stack((H - 0.5 * D, H + 0.5 * D), axis=-1)
                height_bounds = (np.min(H), np.max(H))
                del H, D
            if verbosity > 0:
                print(
                    f"States '{self.name}': Restricting heights to {height_bounds[0]} - {height_bounds[1]} m"
                )
            self._heights = data[cmap[FV.H]].to_numpy()
            if (
                np.min(self._heights) > height_bounds[0]
                or np.max(self._heights) < height_bounds[1]
            ):
                raise ValueError(
                    f"States '{self.name}': Height bounds {height_bounds} m are outside of data height range {np.min(self._heights)} - {np.max(self._heights)} m"
                )
            i0 = 0
            while (
                i0 < len(self._heights) - 1
                and self._heights[i0 + 1] <= height_bounds[0]
            ):
                i0 += 1
            i1 = len(self._heights) - 1
            while i1 > 0 and self._heights[i1 - 1] >= height_bounds[1]:
                i1 -= 1
            if self.sel is None:
                self.sel = {}
            ch = cmap[FV.H]
            self.sel.update({ch: slice(self._heights[i0], self._heights[i1] + 1)})
            self._heights = data[ch].sel({ch: self.sel[ch]}).to_numpy()
            if verbosity > 0:
                print(
                    f"States '{self.name}': Selected {ch} = {self._heights} ({len(self._heights)} heights)"
                )

        # find horizontal bounds:
        if bounds_extra_space is not None:
            assert FV.X in cmap, (
                f"States '{self.name}': x coordinate '{FV.X}' not in cmap {cmap}"
            )
            assert FV.Y in cmap, (
                f"States '{self.name}': y coordinate '{FV.Y}' not in cmap {cmap}"
            )

            # if bounds and self.x_coord is not None and self.x_coord not in self.sel:
            xy_min, xy_max = algo.farm.get_xy_bounds(
                extra_space=bounds_extra_space, algo=algo
            )
            if verbosity > 0:
                print(
                    f"States '{self.name}': Restricting xy to bounds {xy_min} - {xy_max}"
                )
            if self.sel is None:
                self.sel = {}
            for v, i in zip((FV.X, FV.Y), (0, 1)):
                x0, x1 = xy_min[i], xy_max[i]
                x = data[cmap[v]].to_numpy()
                i0 = 0
                while i0 < len(x) - 1 and x[i0 + 1] <= x0:
                    i0 += 1
                i1 = len(x) - 1
                while i1 > 0 and x[i1 - 1] >= x1:
                    i1 -= 1
                if self.sel is None:
                    self.sel = {}
                self.sel.update({cmap[v]: slice(x[i0], x[i1] + 1)})
                if verbosity > 0:
                    hv = data[cmap[v]].sel({cmap[v]: self.sel[cmap[v]]}).to_numpy()
                    print(
                        f"States '{self.name}': Selected {cmap[v]} = {hv[0]} ... {hv[-1]} ({len(hv)} points)"
                    )

    def __preload(
        self,
        algo,
        cmap,
        bounds_extra_space,
        height_bounds,
        verbosity=0,
    ):
        """Helper function for preloading data."""

        assert FC.STATE in cmap, (
            f"States '{self.name}': States coordinate '{FC.STATE}' not in cmap {cmap}"
        )
        states_coord = cmap[FC.STATE]

        if not isinstance(self.data_source, xr.Dataset):
            # check static data:
            fpath = get_input_path(self.data_source)
            if "*" not in str(self.data_source):
                if not fpath.is_file():
                    fpath = StaticData().get_file_path(
                        STATES, fpath.name, check_raw=False
                    )

            # find files:
            prt = fpath.resolve().parent
            glb = fpath.name
            while "*" in str(prt):
                glb = prt.name + "/" + glb
                prt = prt.parent
            files = sorted(list(prt.glob(glb)))
            coords = list(cmap.values())
            vars = [self.var2ncvar.get(v, v) for v in self.variables]

            # pre-process first file:
            fpath = files[0]
            if verbosity > 0:
                print(f"States '{self.name}': Preprocessing first file", fpath.name)
            with xr.open_dataset(fpath, engine=config.nc_engine) as data_first:
                self.drop_vars = [
                    v for v in data_first.data_vars if v not in coords + vars
                ]
                if len(self.drop_vars) > 0 and verbosity > 0:
                    print(f"States '{self.name}': Keeping variables  {vars}")
                    print(f"States '{self.name}': Dropping variables {self.drop_vars}")
                self.preproc_first(
                    algo,
                    data=data_first,
                    cmap=cmap,
                    vars=vars,
                    bounds_extra_space=bounds_extra_space,
                    height_bounds=height_bounds,
                    verbosity=verbosity,
                )
            del data_first

            # read files:
            if verbosity > 0:
                if self.load_mode == "preload":
                    print(
                        f"States '{self.name}': Reading data from '{self.data_source}'"
                    )
                elif self.load_mode == "lazy":
                    print(
                        f"States '{self.name}': Reading header from '{self.data_source}'"
                    )
                else:
                    print(
                        f"States '{self.name}': Reading states from '{self.data_source}'"
                    )

            self.__data_source = get_engine().map(
                _read_nc_file,
                files,
                coords=coords,
                vars=vars,
                nc_engine=config.nc_engine,
                isel=self.isel,
                sel=self.sel,
                minimal=self.load_mode == "fly",
                drop_vars=self.drop_vars,
                check_input_nans=self.check_input_nans,
            )

            if self.load_mode in ["preload", "lazy"]:
                self._input_sizes = [
                    ds.sizes[states_coord] for ds in self.__data_source
                ]
                if self.load_mode == "lazy":
                    try:
                        self.__data_source = [ds.chunk() for ds in self.__data_source]
                    except (ModuleNotFoundError, ValueError) as e:
                        import_module("dask")
                        raise e
                if len(self.__data_source) == 1:
                    self.__data_source = self.__data_source[0]
                else:
                    self.__data_source = xr.concat(
                        self.__data_source,
                        dim=states_coord,
                        coords="minimal",
                        data_vars="minimal",
                        compat="equals",
                        join="exact",
                        combine_attrs="drop",
                    )
                if self.load_mode == "preload":
                    self.__data_source.load()
                self._inds = self.__data_source[states_coord].to_numpy()
                self._N = len(self._inds)

            elif self.load_mode == "fly":
                self._inds = self.__data_source
                self.__data_source = fpath
                self._files_maxi = {f: len(inds) for f, inds in zip(files, self._inds)}
                self._input_sizes = list(self._files_maxi.values())
                self._inds = np.concatenate(self._inds, axis=0)
                self._N = len(self._inds)

            else:
                raise KeyError(
                    f"States '{self.name}': Unknown load_mode '{self.load_mode}', choices: preload, lazy, fly"
                )

            if self.time_format is not None:
                self._inds = pd.to_datetime(
                    self._inds, format=self.time_format
                ).to_numpy()

        # given data is already Dataset:
        else:
            self._inds = self.data_source[states_coord].to_numpy()
            self._N = len(self._inds)

        # make sure state indices are sorted ascending:
        def _is_sorted(a):
            return np.all(a[:-1] <= a[1:])

        if self.check_times and not _is_sorted(self._inds):
            print("\n\nError with state indices, not sorted:\n")
            print(f"State {0:07d}: {self._inds[0]}")
            for i in range(1, self._N):
                print(f"State {i:07d}: {self._inds[i]}")
                if self._inds[i] < self._inds[i - 1]:
                    break
            print()
            raise ValueError(
                f"States '{self.name}': State indices are not sorted ascending: {self._inds[i - 1]} > {self._inds[i]} at position {i - 1}"
            )

        return self.__data_source

    def gen_states_split_size(self):
        """
        Generator for suggested states split sizes for output writing.

        Yields
        ------
        split_size: int or None
            The suggested split size, or None for no splitting

        """
        for size in self._input_sizes:
            yield size

    def load_data(
        self,
        algo,
        cmap,
        variables,
        bounds_extra_space=None,
        height_bounds=None,
        verbosity=0,
    ):
        """
        Load and/or create all model data that is subject to chunking.

        Such data should not be stored under self, for memory reasons. The
        data returned here will automatically be chunked and then provided
        as part of the mdata object during calculations.

        Parameters
        ----------
        algo: foxes.core.Algorithm
            The calculation algorithm
        cmap: dict
            A mapping from foxes variable names to Dataset dimension names
        variables: list of str
            The variables to extract from the Dataset
        bounds_extra_space: float, optional
            The extra space in meters to add to the horizontal wind farm bounds
        height_bounds: tuple, optional
            The (h_min, h_max) height bounds in m. Defaults to H +/- 0.5*D
        verbosity: int
            The verbosity level, 0 = silent

        Returns
        -------
        idata: dict
            The dict has exactly two entries: `data_vars`,
            a dict with entries `name_str -> (dim_tuple, data_ndarray)`;
            and `coords`, a dict with entries `dim_name_str -> dim_array`

        """
        # preload data:
        self.__preload(
            algo,
            cmap,
            bounds_extra_space,
            height_bounds,
            verbosity=verbosity,
        )

        idata = super().load_data(algo, verbosity)

        if self.load_mode == "preload":
            self._coords, data, w = self._get_data(
                self.data_source, cmap, variables, verbosity
            )

            if FC.STATE in self._coords:
                idata["coords"][FC.STATE] = self._coords.pop(FC.STATE)
            else:
                del idata["coords"][FC.STATE]
            if w is not None:
                idata["data_vars"][FV.WEIGHT] = ((FC.STATE,), w)

            vmap = {FC.STATE: FC.STATE}
            self._data_state_keys = []
            self._data_nostate = {}
            for dims, d in data.items():
                dms = tuple([vmap.get(c, self.var(c)) for c in dims])
                if FC.STATE in dims:
                    self._data_state_keys.append(d[0])
                    idata["coords"][dms[-1]] = d[1]
                    idata["data_vars"][d[0]] = (dms, d[2])
                else:
                    self._data_nostate[dims] = (d[1], d[2])
            del data

        return idata

    def set_running(
        self,
        algo,
        data_stash,
        sel=None,
        isel=None,
        verbosity=0,
    ):
        """
        Sets this model status to running, and moves
        all large data to stash.

        The stashed data will be returned by the
        unset_running() function after running calculations.

        Parameters
        ----------
        algo: foxes.core.Algorithm
            The calculation algorithm
        data_stash: dict, optional
            Large data stash, this function adds data here, if given.
            Key: model name. Value: dict, large model data
        sel: dict, optional
            The subset selection dictionary
        isel: dict, optional
            The index subset selection dictionary
        verbosity: int
            The verbosity level, 0 = silent

        """
        super().set_running(algo, data_stash, sel, isel, verbosity)

        if data_stash is not None:
            data_stash[self.name] = dict(
                inds=self._inds,
            )
            if self.load_mode == "preload":
                data_stash[self.name]["data_source"] = self.__data_source
                del self.__data_source
        del self._inds

    def unset_running(
        self,
        algo,
        data_stash,
        sel=None,
        isel=None,
        verbosity=0,
    ):
        """
        Sets this model status to not running, recovering large data
        from stash

        Parameters
        ----------
        algo: foxes.core.Algorithm
            The calculation algorithm
        data_stash: dict, optional
            Reconstruct model data from this stash, if given.
            Key: model name. Value: dict, large model data
        sel: dict, optional
            The subset selection dictionary
        isel: dict, optional
            The index subset selection dictionary
        verbosity: int
            The verbosity level, 0 = silent

        """
        super().unset_running(algo, data_stash, sel, isel, verbosity)

        if data_stash is not None:
            data = data_stash[self.name]
            self._inds = data.pop("inds")

            if self.load_mode == "preload":
                self.__data_source = data.pop("data_source")

    def output_point_vars(self, algo):
        """
        The variables which are being modified by the model.

        Parameters
        ----------
        algo: foxes.core.Algorithm
            The calculation algorithm

        Returns
        -------
        output_vars: list of str
            The output variable names

        """
        return self.ovars

    def size(self):
        """
        The total number of states.

        Returns
        -------
        int:
            The total number of states

        """
        return self._N

    def index(self):
        """
        The index list

        Returns
        -------
        indices: array_like
            The index labels of states, or None for default integers

        """
        if self.running:
            raise ValueError(f"States '{self.name}': Cannot access index while running")
        return self._inds

    def get_calc_data(self, mdata, cmap, variables):
        """
        Gathers data for calculations.

        Call this function from the calculate function of the
        derived class.

        Parameters
        ----------
        mdata: foxes.core.MData
            The mdata object
        cmap: dict
            A mapping from foxes variable names to Dataset dimension names
        variables: list of str
            The variables to extract from the Dataset

        Returns
        -------
        coords: dict
            keys: Foxes variable names, values: 1D coordinate value arrays
        data: dict
            The extracted data, keys are dimension tuples,
            values are tuples (DATA key, variables, data_array)
            where DATA key is the name in the mdata object,
            variables is a list of variable names, and
            data_array is a numpy.ndarray with the data values,
            the last dimension corresponds to the variables
        weights: numpy.ndarray or None
            The weights array, if only state dependent, otherwise
            weights are among data. Shape: (n_states,)

        """
        # prepare
        assert FC.STATE in cmap, (
            f"States '{self.name}': States coordinate '{FC.STATE}' not in cmap {cmap}"
        )
        states_coord = cmap[FC.STATE]
        n_states = mdata.n_states

        # case preload
        if self.load_mode == "preload":
            coords = self._coords
            weights = mdata[FV.WEIGHT] if FV.WEIGHT in mdata else None
            data = deepcopy(self._data_nostate)
            for DATA in self._data_state_keys:
                dims = mdata.dims[DATA]
                vrs = mdata[dims[-1]].tolist()
                dms = tuple(
                    [self.unvar(c) if c != FC.STATE else FC.STATE for c in dims[:-1]]
                    + [dims[-1]]
                )
                data[dms] = (vrs, mdata[DATA].copy())

        # case lazy
        elif self.load_mode == "lazy":
            i0 = mdata.states_i0(counter=True)
            s = slice(i0, i0 + n_states)
            ds = self.data_source.isel({states_coord: s}).load()
            coords, data, weights = self._get_data(ds, cmap, variables, verbosity=0)
            data = {dims: (d[1], d[2]) for dims, d in data.items()}
            del ds

        # case fly
        elif self.load_mode == "fly":
            vars = [self.var2ncvar.get(v, v) for v in variables]
            i0 = mdata.states_i0(counter=True)
            i1 = i0 + n_states
            j0 = 0
            data = []
            for fpath, n in self._files_maxi.items():
                if i0 < j0 or i0 == i1:
                    break
                else:
                    j1 = j0 + n
                    if i0 < j1:
                        a = i0 - j0
                        b = min(i1, j1) - j0
                        assert b > a, (
                            f"States '{self.name}': Invalid state indices for file {fpath}: (i0, i1, j0, j1, a, b) = {(i0, i1, j0, j1, a, b)}"
                        )
                        isel = copy(self.isel) if self.isel is not None else {}
                        isel[states_coord] = slice(a, b)

                        data.append(
                            _read_nc_file(
                                fpath,
                                coords=list(cmap.values()),
                                vars=vars,
                                nc_engine=config.nc_engine,
                                isel=isel,
                                sel=self.sel,
                                minimal=False,
                                drop_vars=self.drop_vars,
                                check_input_nans=self.check_input_nans,
                            )
                        )

                        i0 += b - a
                    j0 = j1

            assert i0 == i1, (
                f"States '{self.name}': Missing states for load_mode '{self.load_mode}': (i0, i1) = {(i0, i1)}"
            )
            if len(data) == 1:
                data = data[0]
            else:
                data = xr.concat(
                    data,
                    dim=states_coord,
                    data_vars="minimal",
                    coords="minimal",
                    compat="override",
                    join="exact",
                    combine_attrs="drop",
                )
            coords, data, weights = self._get_data(data, cmap, variables, verbosity=0)
            data = {dims: (d[1], d[2]) for dims, d in data.items()}

        else:
            raise KeyError(
                f"States '{self.name}': Unknown load_mode '{self.load_mode}', choices: preload, lazy, fly"
            )

        return coords, data, weights

    def interpolate_data(self, idims, icrds, d, pts, tdims, vrs, times):
        """
        Interpolates data to points.

        This function should be implemented in derived classes.

        Parameters
        ----------
        idims: list of str
            The input dimensions, e.g. [state, x, y, height]
        icrds: list of numpy.ndarray
            The input coordinates, each with shape (n_i,)
            where n_i is the number of grid points in dimension i
        d: numpy.ndarray
            The data array, with shape (n1, n2, ..., nv)
            where ni represents the dimension sizes and
            nv is the number of variables
        pts: numpy.ndarray
            The points to interpolate to, with shape (n_pts, n_idims)
        tdims: tuple
            The target dimensions, e.g. (1, m, nv) or (n_states, m, nv)
        vrs: list of str
            The variable names, length nv
        times: numpy.ndarray
            The time coordinates of the states, with shape (n_states,)

        Returns
        -------
        d_interp: numpy.ndarray
            The interpolated data array with shape tdims

        """
        gvars = tuple(icrds)
        try:
            ipars = dict(bounds_error=True, fill_value=None)
            ipars.update(self.interpn_pars)
            d = interpn(gvars, d, pts, **ipars)
        except ValueError as e:
            print(f"\nStates '{self.name}': Interpolation error")
            print(f"INPUT VARS: {idims}")
            print(
                "DATA BOUNDS:",
                [float(np.min(d)) for d in gvars],
                [float(np.max(d)) for d in gvars],
            )
            print(
                "EVAL BOUNDS:",
                [float(np.min(p)) for p in pts.T],
                [float(np.max(p)) for p in pts.T],
            )
            print(
                "\nMaybe you want to try the option 'bounds_error=False' in 'interpn_pars'? This will extrapolate the data.\n"
            )
            raise e

        return d.reshape(tdims)

    def calculate(self, algo, mdata, fdata, tdata):
        """
        The main model calculation.

        This function is executed on a single chunk of data,
        all computations should be based on numpy arrays.

        Parameters
        ----------
        algo: foxes.core.Algorithm
            The calculation algorithm
        mdata: foxes.core.MData
            The model data
        fdata: foxes.core.FData
            The farm data
        tdata: foxes.core.TData
            The target point data

        Returns
        -------
        results: dict
            The resulting data, keys: output variable str.
            Values: numpy.ndarray with shape
            (n_states, n_targets, n_tpoints)

        """
        # prepare
        self.ensure_output_vars(algo, tdata)
        n_states = tdata.n_states
        n_targets = tdata.n_targets
        n_tpoints = tdata.n_tpoints
        points = tdata[FC.TARGETS].reshape(n_states, n_targets * n_tpoints, 3)
        n_pts = points.shape[1]
        times = mdata[FC.STATE]

        # get data for calculation
        coords, data, weights = self.get_calc_data(mdata, self._cmap, self.variables)
        coords[FC.STATE] = np.arange(n_states, dtype=config.dtype_int)

        # interpolate data to points:
        out = {}
        for dims, (vrs, d) in data.items():
            # translate (WD, WS) to (U, V):
            if FV.WD in vrs or FV.WS in vrs:
                assert FV.WD in vrs and (FV.WS in vrs or FV.WS in self.fixed_vars), (
                    f"States '{self.name}': Missing '{FV.WD}' or '{FV.WS}' in data variables {vrs} for dimensions {dims}"
                )
                iwd = vrs.index(FV.WD)
                iws = vrs.index(FV.WS)
                ws = d[..., iws] if FV.WS in vrs else self.fixed_vars[FV.WS]
                d[..., [iwd, iws]] = wd2uv(d[..., iwd], ws, axis=-1)
                del ws

            # prepare points:
            idims = dims[:-1]
            n_vrs = len(vrs)
            tdims = [n_states, n_pts, n_vrs]
            if idims == (FC.STATE, FV.X, FV.Y, FV.H):
                pts = np.append(
                    np.zeros((n_states, n_pts, 1), dtype=config.dtype_double),
                    points,
                    axis=2,
                )
                pts[..., 0] = np.arange(n_states)[:, None]
                pts = pts.reshape(n_states * n_pts, 4)
            elif idims == (FC.STATE, FV.X, FV.Y):
                pts = np.append(
                    np.zeros((n_states, n_pts, 1), dtype=config.dtype_double),
                    points[..., :2],
                    axis=2,
                )
                pts[..., 0] = np.arange(n_states)[:, None]
                pts = pts.reshape(n_states * n_pts, 3)
            elif idims == (FC.STATE, FV.H):
                pts = np.append(
                    np.zeros((n_states, n_pts, 1), dtype=config.dtype_double),
                    points[..., 2, None],
                    axis=2,
                )
                pts[..., 0] = np.arange(n_states)[:, None]
                pts = pts.reshape(n_states * n_pts, 2)
            elif idims == (FC.STATE,):
                if FV.WD in vrs:
                    uv = d[..., [iwd, iws]]
                    d[..., iwd] = uv2wd(uv)
                    d[..., iws] = np.linalg.norm(uv, axis=-1)
                    del uv
                for i, v in enumerate(vrs):
                    if v in self.ovars:
                        out[v] = np.zeros((n_states, n_pts), dtype=config.dtype_double)
                        out[v][:] = d[:, None, i]
                continue
            elif idims == (FV.X, FV.Y, FV.H):
                pts = points[0]
                tdims = (1, n_pts, n_vrs)
            elif idims == (FV.X, FV.Y):
                pts = points[0][:, :2]
                tdims = (1, n_pts, n_vrs)
            elif idims == (FV.H,):
                pts = points[0][:, 2]
                tdims = (1, n_pts, n_vrs)
            else:
                raise ValueError(
                    f"States '{self.name}': Unsupported dimensions {dims} for variables {vrs}"
                )

            # interpolate:
            icrds = [coords[c] for c in idims]
            d = self.interpolate_data(idims, icrds, d, pts, tdims, vrs, times)
            del pts, icrds

            # translate (U, V) into (WD, WS):
            if FV.WD in vrs:
                uv = d[..., [iwd, iws]]
                d[..., iwd] = uv2wd(uv)
                d[..., iws] = np.linalg.norm(uv, axis=-1)
                del uv

            # broadcast if needed:
            if tdims != (n_states, n_pts, n_vrs):
                tmp = d
                d = np.zeros((n_states, n_pts, n_vrs), dtype=config.dtype_double)
                d[:] = tmp
                del tmp

            # set output:
            for i, v in enumerate(vrs):
                out[v] = d[..., i]

        # set fixed variables:
        for v, d in self.fixed_vars.items():
            out[v] = np.full((n_states, n_pts), d, dtype=config.dtype_double)

        # add weights:
        if weights is not None:
            tdata[FV.WEIGHT] = weights[:, None, None]
        elif FV.WEIGHT in out:
            tdata[FV.WEIGHT] = out.pop(FV.WEIGHT).reshape(
                n_states, n_targets, n_tpoints
            )
        else:
            tdata[FV.WEIGHT] = np.full(
                (mdata.n_states, 1, 1), 1 / self._N, dtype=config.dtype_double
            )
        tdata.dims[FV.WEIGHT] = (FC.STATE, FC.TARGET, FC.TPOINT)

        return {v: d.reshape(n_states, n_targets, n_tpoints) for v, d in out.items()}



import numpy as np

from foxes.core import TData
import foxes.variables as FV
import foxes.constants as FC

from .centre import CentreRotor


class DirectMDataInfusion(CentreRotor):
    """
    Direct data infusion of data stored under mdata.

    Attributes
    ----------
    svars2mdvars: dict
        A mapping of state variables to mdata variables.
    mdata_vars: list of str
        The mdata variables to be used for infusion. By default,
        all mdata variables are searched.
    turbine_coord: str, optional
        The mdata coordinate that represents turbine names. By default,
        the second coordinate is used as a candidate if the mdata variable
        has three dimensions.

    :group: models.turbine_types

    """

    def __init__(
        self,
        svars2mdvars=None,
        mdata_vars=None,
        turbine_coord=None,
        **kwargs,
    ):
        """
        Constructor.

        Parameters
        ----------
        svars2mdvars: dict, optional
            A mapping of state variables to mdata variables.
        mdata_vars: list of str, optional
            The mdata variables to be used for infusion. By default,
            all mdata variables are searched.
        turbine_coord: str, optional
            The mdata coordinate that represents turbine names. By default,
            the second coordinate is used as a candidate if the mdata variable
            has three dimensions.
        kwargs: dict, optional
            Additional parameters for RotorModel class

        """
        super().__init__(**kwargs)
        self.svars2mdvars = svars2mdvars
        self.mdata_vars = mdata_vars
        self.turbine_coord = turbine_coord

    def output_farm_vars(self, algo):
        """
        The variables which are being modified by the model.

        Parameters
        ----------
        algo: foxes.core.Algorithm
            The calculation algorithm

        Returns
        -------
        output_vars: list of str
            The output variable names

        """
        if self.svars2mdvars is None:
            self.svars2mdvars = {v: v for v in algo.states.output_point_vars(algo)}

        if self.calc_vars is None:
            vrs = list(self.svars2mdvars.keys())
            assert FV.WEIGHT not in vrs, (
                f"Rotor '{self.name}': svars2mdvars keys {vrs}  contain '{FV.WEIGHT}', please remove"
            )

            if FV.WS in vrs:
                self.calc_vars = [FV.REWS] + [v for v in vrs if v != FV.WS]
            else:
                self.calc_vars = vrs

            if algo.farm_controller.needs_rews2() and FV.REWS2 not in self.calc_vars:
                self.calc_vars.append(FV.REWS2)
            if algo.farm_controller.needs_rews3() and FV.REWS3 not in self.calc_vars:
                self.calc_vars.append(FV.REWS3)

            self.calc_vars = sorted(self.calc_vars)

        if FV.WEIGHT not in self.calc_vars:
            self.calc_vars.append(FV.WEIGHT)
            if FV.WEIGHT not in self.svars2mdvars:
                self.svars2mdvars[FV.WEIGHT] = FV.WEIGHT

        return self.calc_vars

    def calculate(
        self,
        algo,
        mdata,
        fdata,
        rpoints=None,
        rpoint_weights=None,
        store=False,
        downwind_index=None,
    ):
        """
        Calculate ambient rotor effective results.

        Parameters
        ----------
        algo: foxes.core.Algorithm
            The calculation algorithm
        mdata: foxes.core.MData
            The model data
        fdata: foxes.core.FData
            The farm data
        rpoints: numpy.ndarray, optional
            The rotor points, or None for automatic for
            this rotor. Shape: (n_states, n_turbines, n_rpoints, 3)
        rpoint_weights: numpy.ndarray, optional
            The rotor point weights, or None for automatic
            for this rotor. Shape: (n_rpoints,)
        store: bool, optional
            Flag for storing ambient rotor point results
        downwind_index: int, optional
            Only compute for index in the downwind order

        Returns
        -------
        results: dict
            results dict. Keys: Variable name str. Values:
            numpy.ndarray with results, shape: (n_states, n_turbines)

        """
        self.ensure_output_vars(algo, fdata)

        if rpoints is None:
            rpoints = mdata.get(
                FC.ROTOR_POINTS, self.get_rotor_points(algo, mdata, fdata)
            )
        if downwind_index is not None:
            rpoints = rpoints[:, downwind_index, None]
        if rpoint_weights is None:
            rpoint_weights = mdata.get_item(FC.TWEIGHTS, self.rotor_point_weights())

        tdata = TData.from_tpoints(rpoints, rpoint_weights)
        sres = {}
        mdvs = self.mdata_vars if self.mdata_vars is not None else list(mdata.keys())
        for v, w in self.svars2mdvars.items():
            tdata.add(
                v,
                data=np.full_like(rpoints[..., 0], np.nan),
                dims=(FC.STATE, FC.TARGET, FC.TPOINT),
            )

            # check fixed variables
            if hasattr(algo.states, "fixed_vars") and v in algo.states.fixed_vars:
                tdata[v][:] = algo.states.fixed_vars[v]
                sres[v] = tdata[v]
                continue

            # search in mdata variables
            tcoord = self.turbine_coord
            for mdv in mdvs:
                assert mdv in mdata and mdv in mdata.dims, (
                    f"Rotor '{self.name}': mdata variable '{mdv}' not found in mdata {list(mdata.keys())} with dims {list(mdata.dims.keys())}"
                )
                mdat = mdata[mdv]
                dims = mdata.dims[mdv]

                # skip coordinates in the search:
                if dims == (mdv,):
                    continue

                # find variable index in last data array dimension
                vc = dims[-1]
                assert vc in mdata and vc in mdata.dims and mdata.dims[vc] == (vc,), (
                    f"Rotor '{self.name}': mdata coordinate '{vc}' not in mdata or wrong dimensions {mdata.dims}, expected '{(vc,)}'"
                )
                vrs = list(mdata[vc])
                if w in vrs:
                    i = vrs.index(w)
                    mdat = mdat[..., i]
                    dims = dims[:-1]

                    # pure state dependent variable
                    if dims == (FC.STATE,):
                        tdata[v][:] = mdat[:, None, None]

                    # state and turbine dependent variable
                    elif len(dims) == 2 and dims[0] == FC.STATE:
                        assert mdat.shape[1] == mdata.n_turbines, (
                            f"Rotor '{self.name}': mdata variable '{mdv}' has dimensions {dims} and unexpected shape {mdat.shape} for variable '{w}', expected ({mdata.n_states}, {mdata.n_turbines}) for this wind farm with {mdata.n_turbines} turbines"
                        )
                        if tcoord is None:
                            tcoord = dims[1]
                        assert dims[1] == tcoord, (
                            f"Rotor '{self.name}': mdata variable '{mdv}' has unexpected dimensions {dims} for variable '{w}', expected ({FC.STATE}, {tcoord})"
                        )

                        tdata[v][:] = mdat[:, :, None]

                    else:
                        if tcoord is None:
                            tcoord = "<turbine>"
                        raise ValueError(
                            f"Rotor '{self.name}': mdata variable '{mdv}' has unexpected dimensions {dims} for variable '{w}' at position {i}, expected ({FC.STATE},) or ({FC.STATE}, {tcoord})"
                        )

                    sres[v] = tdata[v]
                    break

            if v not in sres:
                raise ValueError(
                    f"Rotor '{self.name}': mdata variable '{w}' not found in any of the mdata variables {mdvs}"
                )

        if store:
            algo.add_to_chunk_store(
                FC.ROTOR_POINTS,
                rpoints,
                dims=(FC.STATE, FC.TURBINE, FC.ROTOR_POINT, FC.XYH),
                mdata=mdata,
            )
            algo.add_to_chunk_store(
                FC.ROTOR_WEIGHTS,
                rpoint_weights,
                dims=(FC.ROTOR_POINT,),
                mdata=mdata,
            )
            algo.add_to_chunk_store(
                FC.AMB_ROTOR_RES,
                sres,
                dims=(FC.STATE, FC.TURBINE, FC.ROTOR_POINT),
                mdata=mdata,
            )
            algo.add_to_chunk_store(
                FC.WEIGHT_RES,
                tdata[FV.WEIGHT],
                dims=(FC.STATE, FC.TURBINE, FC.ROTOR_POINT),
                mdata=mdata,
            )

        self.eval_rpoint_results(
            algo,
            mdata,
            fdata,
            tdata,
            rpoint_weights,
            downwind_index,
            copy_to_ambient=True,
        )

        return {v: fdata[v] for v in self.output_farm_vars(algo)}



import matplotlib.pyplot as plt
import numpy as np

import foxes.constants as FC

from .output import Output


class StateTurbineMap(Output):
    """
    Creates heat maps with turbines on one axis
    and states on the other axis.

    Attributes
    ----------
    results: xarray.Dataset
        The farm results

    :group: output

    """

    def __init__(self, farm_results, **kwargs):
        """
        Constructor.

        Parameters
        ----------
        farm_results: xarray.Dataset
            The farm results
        kwargs: dict, optional
            Additional parameters for the base class

        """
        super().__init__(**kwargs)
        self.results = farm_results

    def plot_map(
        self,
        variable,
        title=None,
        cbar_label=None,
        ax=None,
        figsize=None,
        rotate_xlabels=None,
        **kwargs,
    ):
        """
        Plot the heat map for the selected variable.

        Parameters
        ----------
        variable: str
            The variable to plot
        title: str, optional
            The plot title
        ax: pyplot.Axis, optional
            The axis
        figsize: tuple
            The figsize argument for plt.subplots()
            in case ax is not provided
        rotate_xlabels: float, optional
            Rotate the x-labels by this number of degrees
        kwargs: dict, optional
            Additional parameters for plt.pcolormesh()

        Returns
        -------
        ax: pyplot.Axis
            The plot axis

        """
        if self.nofig:
            return None

        turbines = self.results[FC.TURBINE].to_numpy()
        states = self.results[FC.STATE].to_numpy()

        if ax is None:
            __, ax = plt.subplots(figsize=figsize)
        fig = ax.get_figure()

        ds = states[-1] - states[-2]
        states = np.append(states, states[-1] + ds)
        turbines = np.arange(len(turbines) + 1)

        y, x = np.meshgrid(turbines, states)
        z = self.results[variable].to_numpy()

        prgs = {"shading": "flat"}
        prgs.update(kwargs)

        c = ax.pcolormesh(x, y, z, **prgs)

        ax.set_yticks(turbines[:-1] + 0.5)
        ax.set_yticklabels(turbines[:-1])
        xt = np.asarray(ax.get_xticks())
        xtl = ax.get_xticklabels()
        try:
            xt, ar = np.unique(xt.astype(int), return_index=True)
            xtl = [int(float(xtl[i].get_text())) for i in ar]
        except ValueError:
            pass
        ax.set_xticks(
            xt[:-1] + 0.5 * (xt[-1] - xt[-2]), xtl[:-1], rotation=rotate_xlabels
        )
        if len(turbines) > 10:
            yt = ax.get_yticks()
            ytl = [None for t in yt]
            ytl[::5] = ax.get_yticklabels()[::5]
            ax.set_yticks(yt, ytl)
        fig.colorbar(c, ax=ax, label=cbar_label)

        ax.set_title(title)
        ax.set_ylabel("Turbine index")
        ax.set_xlabel("State")

        return ax



import argparse
import numpy as np
import pandas as pd
import matplotlib.pyplot as plt

import foxes
import foxes.variables as FV


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "-nt", "--n_t", help="The number of turbines", type=int, default=5
    )
    parser.add_argument(
        "-s",
        "--states",
        help="The timeseries input file (path or static)",
        default="states.csv",
    )
    parser.add_argument(
        "-t",
        "--turbine_file",
        help="The P-ct-curve csv file (path or static)",
        default="NREL-5MW-D126-H90.csv",
    )
    parser.add_argument(
        "-dx", "--deltax", help="The turbine distance in x", type=float, default=500.0
    )
    parser.add_argument(
        "-dy", "--deltay", help="Turbine layout y step", type=float, default=0.0
    )
    parser.add_argument("-r", "--rotor", help="The rotor model", default="centre")
    parser.add_argument(
        "-p", "--pwakes", help="The partial wakes models", default=None, nargs="+"
    )
    parser.add_argument(
        "-f", "--pmax_file", help="The max_P csv file", default="power_mask.csv"
    )
    parser.add_argument(
        "-w",
        "--wakes",
        help="The wake models",
        default=["Jensen_linear_k007"],
        nargs="+",
    )
    parser.add_argument(
        "-m", "--tmodels", help="The turbine models", default=[], nargs="+"
    )
    parser.add_argument(
        "-sl",
        "--show_layout",
        help="Flag for showing layout figure",
        action="store_true",
    )
    parser.add_argument("-e", "--engine", help="The engine", default="process")
    parser.add_argument(
        "-n", "--n_cpus", help="The number of cpus", default=None, type=int
    )
    parser.add_argument(
        "-c",
        "--chunksize_states",
        help="The chunk size for states",
        default=None,
        type=int,
    )
    parser.add_argument(
        "-C",
        "--chunksize_points",
        help="The chunk size for points",
        default=None,
        type=int,
    )
    parser.add_argument(
        "-nf", "--nofig", help="Do not show figures", action="store_true"
    )
    args = parser.parse_args()

    print("\nReading file", args.pmax_file)
    cols = [f"Pmax_{i}" for i in range(args.n_t)]
    Pmax_data = pd.read_csv(args.pmax_file, usecols=cols).to_numpy()

    mbook = foxes.models.ModelBook()
    ttype = foxes.models.turbine_types.PCtFile(args.turbine_file)
    mbook.turbine_types[ttype.name] = ttype
    mbook.turbine_models["set_Pmax"] = foxes.models.turbine_models.SetFarmVars()
    mbook.turbine_models["set_Pmax"].add_var(FV.MAX_P, Pmax_data)
    models = args.tmodels + ["set_Pmax", ttype.name, "PMask"]

    with foxes.Engine.new(
        engine_type=args.engine,
        n_procs=args.n_cpus,
        chunk_size_states=args.chunksize_states,
        chunk_size_points=args.chunksize_points,
    ):
        if not args.nofig:
            fig, axs = plt.subplots(1, 2, figsize=(10, 4))
            o = foxes.output.TurbineTypeCurves(mbook)
            o.plot_curves(ttype.name, [FV.P, FV.CT], axs=axs, P_max=3000.0)
            plt.show()
            plt.close(fig)

            """
            # TODO: ct needs fix
            fig, axs = plt.subplots(1, 2, figsize=(10, 4))
            o = foxes.output.TurbineTypeCurves(mbook)
            o.plot_curves(ttype.name, [FV.P, FV.CT], axs=axs, P_max=6000.0)
            plt.show()
            plt.close(fig)
            """

        states = foxes.input.states.StatesTable(
            data_source=args.states,
            output_vars=[FV.WS, FV.WD, FV.TI, FV.RHO],
            fixed_vars={FV.WD: 270.0, FV.TI: 0.05, FV.RHO: 1.225},
        )

        farm = foxes.WindFarm()
        foxes.input.farm_layout.add_row(
            farm=farm,
            xy_base=[0.0, 0.0],
            xy_step=[args.deltax, args.deltay],
            n_turbines=args.n_t,
            turbine_models=models,
        )

        if not args.nofig and args.show_layout:
            ax = foxes.output.FarmLayoutOutput(farm).get_figure()
            plt.show()
            plt.close(ax.get_figure())

        algo = foxes.algorithms.Downwind(
            farm,
            states,
            rotor_model=args.rotor,
            wake_models=args.wakes,
            wake_frame="rotor_wd",
            partial_wakes=args.pwakes,
            mbook=mbook,
            verbosity=0,
        )

        outputs = [
            FV.D,
            FV.WD,
            FV.AMB_REWS,
            FV.REWS,
            FV.AMB_P,
            FV.P,
            FV.CT,
            FV.WEIGHT,
            FV.MAX_P,
        ]

        # run calculation with power mask:
        farm_results = algo.calc_farm(outputs=outputs)

        fr = farm_results.to_dataframe()
        print(fr[[FV.WD, FV.AMB_REWS, FV.REWS, FV.AMB_P, FV.P, FV.CT, FV.MAX_P]])

        o = foxes.output.FarmResultsEval(farm_results)
        P0 = o.calc_mean_farm_power(ambient=True)
        P = o.calc_mean_farm_power()
        print(f"\nFarm power: {P / 1000:.1f} MW, Efficiency = {P / P0 * 100:.2f} %")

        o1 = foxes.output.StateTurbineMap(farm_results)

        # run calculation without power mask:
        for t in farm.turbines:
            t.models.remove("set_Pmax")
            t.models.remove("PMask")

        farm_results = algo.calc_farm(outputs=outputs)

        fr = farm_results.to_dataframe()
        print(fr[[FV.WD, FV.AMB_REWS, FV.REWS, FV.AMB_P, FV.P, FV.CT, FV.MAX_P]])

        o = foxes.output.FarmResultsEval(farm_results)
        P0 = o.calc_mean_farm_power(ambient=True)
        P = o.calc_mean_farm_power()
        print(f"\nFarm power: {P / 1000:.1f} MW, Efficiency = {P / P0 * 100:.2f} %")

        if not args.nofig:
            o0 = foxes.output.StateTurbineMap(farm_results)

            # show power:
            fig, axs = plt.subplots(1, 3, figsize=(11, 5))
            o0.plot_map(
                FV.P,
                ax=axs[0],
                edgecolor="white",
                title="Power, no power mask",
                cmap="YlOrRd",
                vmin=0,
                vmax=np.nanmax(Pmax_data),
            )
            o1.plot_map(
                FV.MAX_P,
                ax=axs[1],
                edgecolor="white",
                cmap="YlOrRd",
                title="Power mask",
                vmin=0,
                vmax=np.nanmax(Pmax_data),
            )
            o1.plot_map(
                FV.P,
                ax=axs[2],
                edgecolor="white",
                cmap="YlOrRd",
                title="Power, with power mask",
                vmin=0,
                vmax=np.nanmax(Pmax_data),
            )
            plt.show()
            plt.close(fig)

            # show ct:
            fig, axs = plt.subplots(1, 3, figsize=(11, 5))
            o0.plot_map(
                FV.CT,
                ax=axs[0],
                edgecolor="white",
                title="ct, no power mask",
                cmap="YlGn",
                vmin=0,
                vmax=1.0,
            )
            o1.plot_map(
                FV.MAX_P,
                ax=axs[1],
                edgecolor="white",
                cmap="YlOrRd",
                title="Power mask",
                vmin=0,
                vmax=np.nanmax(Pmax_data),
            )
            o1.plot_map(
                FV.CT,
                ax=axs[2],
                edgecolor="white",
                cmap="YlGn",
                title="ct, with power mask",
                vmin=0,
                vmax=1.0,
            )
            plt.show()



import numpy as np
from xarray import Dataset, SerializationWarning
from pathlib import Path
import warnings

import foxes.variables as FV


def compute_scale_and_offset(min, max, n, hasnan=True):
    """
    Computes scale_factor and add_offset for packing data
    into n-bit integers.

    Parameters
    ----------
    min: float
        Minimum value of the data
    max: float
        Maximum value of the data
    n: int
        Number of bits for packing
    hasnan: bool
        NaN present in the data

    Returns
    -------
    scale_factor: float
        The scale factor
    add_offset: float
        The add offset
    fill_value: float
        The fill value for NaN

    Notes
    -----
    Source: https://docs.unidata.ucar.edu/nug/current/best_practices.html

    """
    if min == max:
        max = min + 1
    if hasnan:
        scale_factor = (max - min) / (2**n - 2)
        add_offset = 0.5 * (max + min)
        fill_value = -(2 ** (n - 1))
    else:
        scale_factor = (max - min) / (2**n - 1)
        add_offset = min + 2 ** (n - 1) * scale_factor
        fill_value = None
    return scale_factor, add_offset, fill_value


def pack_value(unpacked_value, scale_factor, add_offset, dtype, fill_value):
    """
    Pack a floating point value into an integer representation.

    Parameters
    ----------
    unpacked_value: float or np.ndarray
        The floating point value(s) to be packed
    scale_factor: float
        The scale factor
    add_offset: float
        The add offset
    dtype: numpy.dtype
        The dtype of packed values
    fill_value: float
        The fill value for NaN

    Returns
    -------
    packed_value: int or np.ndarray
        The packed integer value(s)

    :group: utils

    """
    if fill_value is None:
        return np.floor((unpacked_value - add_offset) / scale_factor).astype(dtype)
    else:
        return np.where(
            np.isnan(unpacked_value),
            fill_value,
            np.floor((unpacked_value - add_offset) / scale_factor),
        ).astype(dtype)


def unpack_value(packed_value, scale_factor, add_offset, fill_value):
    """
    Unpack an integer representation back into a floating point value.

    Parameters
    ----------
    packed_value: int or np.ndarray
        The packed integer value(s) to be unpacked
    scale_factor: float
        The scale factor
    add_offset: float
        The add offset
    fill_value: float
        The fill value for NaN

    Returns
    -------
    unpacked_value: float or np.ndarray
        The unpacked floating point value(s)

    :group: utils

    """
    if fill_value is None:
        return (packed_value * scale_factor + add_offset).astype(scale_factor.dtype)
    else:
        return np.where(
            packed_value == fill_value, np.nan, packed_value * scale_factor + add_offset
        ).astype(scale_factor.dtype)


def get_encoding(data, complevel=5):
    """
    Get the encoding parameters for a numpy array.

    Parameters
    ----------
    data: np.ndarray
        The numpy array for which to get the encoding information.
    complevel: int
        The compression level (1-9)

    Returns
    -------
    encoding: dict
        The encoding information of the numpy array.

    :group: utils

    """
    enc = {"zlib": True, "complevel": complevel}
    if np.issubdtype(data.dtype, np.integer):
        for t in [np.int8, np.uint8, np.int16, np.uint16, np.int32, np.uint32]:
            if np.all(data == data.astype(t)):
                enc["dtype"] = t.__name__
    elif np.issubdtype(data.dtype, np.floating):
        min = np.min(data)
        max = np.max(data)
        hasnan = np.any(np.isnan(data))
        for t, n in zip([np.int8, np.int16], [8, 16]):
            scale_factor, add_offset, fill_value = compute_scale_and_offset(
                min, max, n, hasnan
            )
            packed = pack_value(data, scale_factor, add_offset, t, fill_value)
            unpacked = unpack_value(packed, scale_factor, add_offset, fill_value)
            try:
                np.testing.assert_allclose(data, unpacked, atol=scale_factor)
                enc["dtype"] = t.__name__
                enc["scale_factor"] = scale_factor
                enc["add_offset"] = add_offset
                enc["_FillValue"] = fill_value
                break
            except AssertionError:
                continue
    return enc


def write_nc(
    ds,
    fpath,
    round={},
    complevel=5,
    nc_engine="netcdf4",
    verbosity=1,
    **kwargs,
):
    """
    Writes a dataset to netCDF file

    Parameters
    ----------
    fpath: str
        Path to the output file, should be nc
    round: dict or int
        The rounding digits, falling back to defaults
        if variable not found. If int, applies to all variables.
    complevel: int
        The compression level
    nc_engine: str
        The NetCDF engine to use
    verbosity: int
        The verbosity level, 0 = silent
    kwargs: dict, optional
        Additional parameters for xarray.to_netcdf

    :group: utils

    """

    def _round(x, v, d):
        """Helper function to round values"""
        if d is not None:
            if np.issubdtype(x.dtype, np.integer):
                return x
            elif np.issubdtype(x.dtype, np.floating):
                if verbosity > 1:
                    print(f"File {fpath.name}: Rounding {v} to {d} decimals")
                r = np.round(x, d)
                return r
        return x

    enc = {}
    fpath = Path(fpath)
    if round is not None:
        crds = {}
        for v, x in ds.coords.items():
            if isinstance(round, int):
                d = round
            else:
                d = round.get(v, FV.get_default_digits(v))
            crds[v] = _round(x.to_numpy(), v, d)
            enc[v] = get_encoding(crds[v], complevel=complevel)
            # print("WRITENC ENC",v, enc[v])
        dvrs = {}
        for v, x in ds.data_vars.items():
            if isinstance(round, int):
                d = round
            else:
                d = round.get(v, FV.get_default_digits(v))
            if v != FV.WEIGHT:
                dvrs[v] = (x.dims, _round(x.to_numpy(), v, d))
            else:
                dvrs[v] = (x.dims, x.to_numpy())
            enc[v] = get_encoding(dvrs[v][1], complevel=complevel)
            # print("WRITENC ENC",v, enc[v])
        ds = Dataset(coords=crds, data_vars=dvrs)

    if verbosity > 0:
        print("Writing file", fpath)

    kw = dict(encoding=enc, engine=nc_engine)
    kw.update(kwargs)

    # silencing a warning about _FillValue = None
    with warnings.catch_warnings():
        warnings.simplefilter("ignore", category=SerializationWarning)
        ds.to_netcdf(fpath, **kw)



import numpy as np
from scipy.interpolate import interpn

from foxes.core.states import States
from foxes.config import config, get_input_path
from foxes.utils import ReaderWRG, weibull_weights
from foxes.data import STATES
import foxes.variables as FV
import foxes.constants as FC


class WRGStates(States):
    """
    Ambient states based on WRG data

    Attributes
    ----------
    wrg_fname: str
        Name of the WRG file
    ws_bins: numpy.ndarray
        The wind speed bins, including
        lower and upper bounds, shape: (n_ws_bins+1,)
    fixed_vars: dict
        Fixed uniform variable values, instead of
        reading from data
    bounds_extra_space: float or str
        The extra space, either float in m,
        or str for units of D, e.g. '2.5D'
    interpn_pars: dict
        Additional parameters for scipy.interpolate.interpn

    :group: input.states

    """

    def __init__(
        self,
        wrg_fname,
        ws_bins,
        fixed_vars={},
        bounds_extra_space="1D",
        **interpn_pars,
    ):
        """
        Constructor

        Parameters
        ----------
        wrg_fname: str
            Name of the WRG file
        ws_bins: list of float
            The wind speed bins, including
            lower and upper bounds
        fixed_vars: dict
            Fixed uniform variable values, instead of
            reading from data
        bounds_extra_space: float or str, optional
            The extra space, either float in m,
            or str for units of D, e.g. '2.5D'
        interpn_pars: dict, optional
            Additional parameters for scipy.interpolate.interpn

        """
        super().__init__()
        self.wrg_fname = wrg_fname
        self.ws_bins = np.asarray(ws_bins)
        self.fixed_vars = fixed_vars
        self.bounds_extra_space = bounds_extra_space
        self.interpn_pars = interpn_pars

    def load_data(self, algo, verbosity=0):
        """
        Load and/or create all model data that is subject to chunking.

        Such data should not be stored under self, for memory reasons. The
        data returned here will automatically be chunked and then provided
        as part of the mdata object during calculations.

        Parameters
        ----------
        algo: foxes.core.Algorithm
            The calculation algorithm
        verbosity: int
            The verbosity level, 0 = silent

        Returns
        -------
        idata: dict
            The dict has exactly two entries: `data_vars`,
            a dict with entries `name_str -> (dim_tuple, data_ndarray)`;
            and `coords`, a dict with entries `dim_name_str -> dim_array`

        """
        # read wrg file:
        fpath = get_input_path(self.wrg_fname)
        if not fpath.is_file():
            if verbosity > 0:
                print(
                    f"States '{self.name}': Reading static data '{self.wrg_fname}' from context '{STATES}'"
                )
            fpath = algo.dbook.get_file_path(STATES, self.wrg_fname, check_raw=False)
            if verbosity > 0:
                print(f"Path: {fpath}")
        elif verbosity:
            print(f"States '{self.name}': Reading file {fpath}")
        wrg = ReaderWRG(fpath)
        p0 = np.array([wrg.x0, wrg.y0], dtype=config.dtype_double)
        nx = wrg.nx
        ny = wrg.ny
        ns = wrg.n_sectors
        res = wrg.resolution
        p1 = p0 + np.array([nx * res, ny * res])
        if verbosity > 0:
            print(f"States '{self.name}': Data bounds {p0} - {p1}")

        # find bounds:
        if self.bounds_extra_space is not None:
            xy_min, xy_max = algo.farm.get_xy_bounds(
                extra_space=self.bounds_extra_space, algo=algo
            )
            if verbosity > 0:
                print(f"States '{self.name}': Farm bounds {xy_min} - {xy_max}")
            ij_min = np.asarray((xy_min - p0) / res, dtype=config.dtype_int)
            ij_max = np.asarray((xy_max - p0) / res, dtype=config.dtype_int) + 1
            sx = slice(ij_min[0], ij_max[0])
            sy = slice(ij_min[1], ij_max[1])
        else:
            sx = np.s_[:]
            sy = np.s_[:]
        self._x = p0[0] + np.arange(nx) * res
        self._x = self._x[sx]
        self._y = p0[1] + np.arange(ny) * res
        self._y = self._y[sy]
        if len(self._x) < 2 or len(self._y) < 2:
            raise ValueError("No overlap between data bounds and farm bounds")
        p0[0] = np.min(self._x)
        p0[1] = np.min(self._y)
        p1[0] = np.max(self._x)
        p1[1] = np.max(self._y)
        if verbosity > 0:
            print(f"States '{self.name}':  New bounds {p0} - {p1}")

        # store data:
        A = []
        k = []
        f = []
        for s in range(ns):
            A.append(wrg.data[f"As_{s}"].to_numpy().reshape(ny, nx)[sy, sx])
            k.append(wrg.data[f"Ks_{s}"].to_numpy().reshape(ny, nx)[sy, sx])
            f.append(wrg.data[f"fs_{s}"].to_numpy().reshape(ny, nx)[sy, sx])
        del wrg
        A = np.stack(A, axis=0).T
        k = np.stack(k, axis=0).T
        f = np.stack(f, axis=0).T
        self._data = np.stack([A, k, f], axis=-1)  # (x, y, wd, AKfs)

        # store ws and wd:
        self.VARS = self.var("VARS")
        self.DATA = self.var("DATA")
        self._wds = np.arange(0.0, 360.0, 360 / ns)
        self._wsd = self.ws_bins[1:] - self.ws_bins[:-1]
        self._wss = 0.5 * (self.ws_bins[:-1] + self.ws_bins[1:])
        self._N = len(self._wss) * ns
        data = np.zeros((len(self._wss), ns, 3), dtype=config.dtype_double)
        data[..., 0] = self._wss[:, None]
        data[..., 1] = self._wds[None, :]
        data[..., 2] = self._wsd[:, None]
        data = data.reshape(self._N, 3)
        idata = super().load_data(algo, verbosity)
        idata["coords"][self.VARS] = ["ws", "wd", "dws"]
        idata["data_vars"][self.DATA] = ((FC.STATE, self.VARS), data)

        return idata

    def size(self):
        """
        The total number of states.

        Returns
        -------
        int:
            The total number of states

        """
        return self._N

    def output_point_vars(self, algo):
        """
        The variables which are being modified by the model.

        Parameters
        ----------
        algo: foxes.core.Algorithm
            The calculation algorithm

        Returns
        -------
        output_vars: list of str
            The output variable names

        """
        ovars = set([FV.WS, FV.WD])
        ovars.update(self.fixed_vars.keys())
        return list(ovars)

    def calculate(self, algo, mdata, fdata, tdata):
        """
        The main model calculation.

        This function is executed on a single chunk of data,
        all computations should be based on numpy arrays.

        Parameters
        ----------
        algo: foxes.core.Algorithm
            The calculation algorithm
        mdata: foxes.core.MData
            The model data
        fdata: foxes.core.FData
            The farm data
        tdata: foxes.core.TData
            The target point data

        Returns
        -------
        results: dict
            The resulting data, keys: output variable str.
            Values: numpy.ndarray with shape
            (n_states, n_targets, n_tpoints)

        """
        # prepare:
        self.ensure_output_vars(algo, tdata)
        n_states = tdata.n_states
        n_targets = tdata.n_targets
        n_tpoints = tdata.n_tpoints
        n_pts = n_states * n_targets * n_tpoints
        points = tdata[FC.TARGETS]
        ws = mdata[self.DATA][:, 0]
        wd = mdata[self.DATA][:, 1]
        wsd = mdata[self.DATA][:, 2]

        out = {}

        out[FV.WS] = tdata[FV.WS]
        out[FV.WS][:] = ws[:, None, None]

        out[FV.WD] = tdata[FV.WD]
        out[FV.WD][:] = wd[:, None, None]

        for v, d in self.fixed_vars.items():
            out[v] = tdata[v]
            out[v][:] = d

        # interpolate A, k, f from x, y, wd
        z = points[..., 2].copy()
        points[..., 2] = wd[:, None, None]
        pts = points.reshape(n_pts, 3)
        gvars = (self._x, self._y, self._wds)
        try:
            ipars = dict(bounds_error=True, fill_value=None)
            ipars.update(self.interpn_pars)
            data = interpn(gvars, self._data, pts, **ipars).reshape(
                n_states, n_targets, n_tpoints, 3
            )
        except ValueError as e:
            print(f"\nStates '{self.name}': Interpolation error")
            print("INPUT VARS: (x, y, wd)")
            print(
                "DATA BOUNDS:",
                [float(np.min(d)) for d in gvars],
                [float(np.max(d)) for d in gvars],
            )
            print(
                "EVAL BOUNDS:",
                [float(np.min(p)) for p in pts.T],
                [float(np.max(p)) for p in pts.T],
            )
            print(
                "\nMaybe you want to try the option 'bounds_error=False'? This will extrapolate the data.\n"
            )
            raise e

        A = data[..., 0]
        k = data[..., 1]
        f = data[..., 2]
        points[..., 2] = z
        del data, gvars, pts, z, wd

        tdata.add(
            FV.WEIGHT,
            f,
            dims=(FC.STATE, FC.TARGET, FC.TPOINT),
        )

        tdata[FV.WEIGHT] *= weibull_weights(
            ws=out[FV.WS],
            ws_deltas=wsd[:, None, None],
            A=A,
            k=k,
        )

        return out



import pandas as pd
from pathlib import Path
import inspect

import foxes
import foxes.variables as FV


def test():
    thisdir = Path(inspect.getabsfile(inspect.currentframe())).parent
    print("TESTDIR:", thisdir)

    cfile = thisdir / "flappy" / "results.csv.gz"
    tfile = thisdir / "NREL-5MW-D126-H90.csv"
    sfile = thisdir / "states.csv.gz"
    lfile = thisdir / "test_farm.csv"

    mbook = foxes.models.ModelBook()
    ttype = foxes.models.turbine_types.PCtFile(
        data_source=tfile, var_ws_ct=FV.REWS, var_ws_P=FV.REWS
    )
    mbook.turbine_types[ttype.name] = ttype

    states = foxes.input.states.StatesTable(
        data_source=sfile,
        output_vars=[FV.WS, FV.WD, FV.TI, FV.RHO, FV.MOL],
        var2col={FV.WS: "ws", FV.WD: "wd", FV.TI: "ti", FV.MOL: "mol"},
        fixed_vars={FV.RHO: 1.225, FV.Z0: 0.05, FV.H: 100.0},
        profiles={FV.WS: "ABLLogWsProfile"},
    )

    farm = foxes.WindFarm()
    foxes.input.farm_layout.add_from_file(
        farm,
        lfile,
        col_x="x",
        col_y="y",
        col_H="H",
        turbine_models=[ttype.name],
        verbosity=0,
    )

    algo = foxes.algorithms.Downwind(
        farm,
        states,
        mbook=mbook,
        rotor_model="centre",
        wake_models=["Jensen_linear_k007"],
        wake_frame="rotor_wd",
        partial_wakes={"Jensen_linear_k007": "centre"},
        verbosity=0,
    )

    with foxes.Engine.new("threads", chunk_size_states=2000):
        data = algo.calc_farm()

    df = data.to_dataframe()[[FV.AMB_WD, FV.WD, FV.AMB_REWS, FV.REWS, FV.AMB_P, FV.P]]

    print()
    print("TRESULTS\n")
    print(df)

    print("\nReading file", cfile)
    fdata = pd.read_csv(cfile)
    print(fdata)

    print("\nVERIFYING\n")
    df[FV.WS] = df["REWS"]
    df[FV.AMB_WS] = df["AMB_REWS"]

    delta = df.reset_index() - fdata
    print(delta)
    print(delta.max())
    chk = delta[[FV.AMB_WS, FV.AMB_P, FV.WS, FV.P]].abs()
    print(chk.max())

    assert (chk[FV.AMB_WS] < 1e-5).all()
    assert (chk[FV.AMB_P] < 1e-3).all()
    assert (chk[FV.WS] < 1e-5).all()
    assert (chk[FV.P] < 1e-3).all()


if __name__ == "__main__":
    test()



from abc import ABCMeta, abstractmethod
import numpy as np

from foxes.utils import delta_wd
import foxes.variables as FV
import foxes.constants as FC


class ConvCrit(metaclass=ABCMeta):
    """
    Abstract base class for convergence criteria

    Attributes
    ----------
    name: str, optional
        The convergence criteria name

    :group: algorithms.iterative.models

    """

    def __init__(self, name=None):
        """
        Constructor.

        Parameters
        ----------
        name: str, optional
            The convergence criteria name

        """
        self.name = name if name is not None else type(self).__name__

        self._deltas = None
        self._conv_states = None
        self.__no_subs = False

    def disable_subsets(self, no_subs=True):
        """
        Disable subset state selection in iterative algorithm.

        This is needed if the convergence criterion requires
        all states to be calculated in each iteration.

        Parameters
        ----------
        no_subs: bool
            Disable subsets flag

        """
        self.__no_subs = no_subs

    @property
    def no_subs(self):
        """
        Get the disable subsets flag.

        Returns
        -------
        no_subs: bool
            Disable subsets flag

        """
        return self.__no_subs

    @abstractmethod
    def check_converged(self, algo, prev_results, results, verbosity=0):
        """
        Check convergence criteria.

        Parameters
        ----------
        algo: foxes.core.Algorithm
            The calculation algorithm
        prev_results: xarray.Dataset
            The farm results of previous
            iteration, or None if first
        results: xarray.Dataset
            The farm results of current
            iteration
        verbosity: int
            The verbosity level, 0 = silent

        Returns
        -------
        convergence: bool
            Convergence flag, true if converged

        """
        pass

    @property
    def deltas(self):
        """
        Get the most recent evaluation deltas.

        Returns
        -------
        deltas: dict
            The most recent evaluation deltas

        """
        return self._deltas

    @property
    def conv_states(self):
        """
        Get the convergence state per state.

        Returns
        -------
        conv_states: numpy.ndarray, bool
            The convergence state per state

        """
        return None if self.no_subs else self._conv_states


class ConvCritList(ConvCrit):
    """
    A list of convergence criteria

    Attributes
    ----------
    crits: list of ConvCrit
        The criteria

    :group: algorithms.iterative.models

    """

    def __init__(self, crits=[], name=None):
        """
        Constructor.

        Parameters
        ----------
        crits: list of ConvCrit
            The criteria
        name: str, optional
            The convergence criteria name

        """
        super().__init__(name)
        self.crits = crits

    def add_crit(self, crit):
        """
        Add a convergence criterion

        Parameters
        ----------
        crit: ConvCrit
            The criterion

        """
        self.crits.append(crit)

    def check_converged(self, algo, prev_results, results, verbosity=0):
        """
        Check convergence criteria.

        Parameters
        ----------
        algo: foxes.core.Algorithm
            The calculation algorithm
        prev_results: xarray.Dataset
            The farm results of previous
            iteration, or None if first
        results: xarray.Dataset
            The farm results of current
            iteration
        verbosity: int
            The verbosity level, 0 = silent

        Returns
        -------
        convergence: bool
            Convergence flag, true if converged

        """
        self._failed = None
        self._conv_states = None
        for c in self.crits:
            conv = c.check_converged(algo, prev_results, results, verbosity)

            if self._conv_states is None:
                self._conv_states = c.conv_states
                self._deltas = c.deltas
            else:
                self._conv_states = self._conv_states & c.conv_states
                self._deltas = {v: max(self._deltas[v], d) for v, d in c.deltas.items()}

            if self.failed is None and not conv:
                self._failed = c

        return self._failed is None


class ConvVarDelta(ConvCrit):
    """
    Requires convergence of a selection of variables.

    Attributes
    ----------
    limits: dict
        The convergence limits. Keys: variables str,
        values: float values
    wd_vars: list of str
        The wind direction type variables (unit deg)

    :group: algorithms.iterative.models

    """

    def __init__(self, limits, wd_vars=None, name=None):
        """
        Constructor.

        Parameters
        ----------
        limits: dict
            The convergence limits. Keys: variables str,
            values: float values
        wd_vars: list of str, optional
            The wind direction type variables (unit deg)
        name: str, optional
            The convergence criteria name

        """
        super().__init__(name)
        self.limits = limits
        if wd_vars is None:
            self.wd_vars = [FV.WD, FV.AMB_WD, FV.YAW, FV.AMB_YAW]
        else:
            self.wd_vars = wd_vars

    def check_converged(self, algo, prev_results, results, verbosity=0):
        """
        Check convergence criteria.

        Parameters
        ----------
        algo: foxes.core.Algorithm
            The calculation algorithm
        prev_results: xarray.Dataset
            The farm results of previous
            iteration, or None if first
        results: xarray.Dataset
            The farm results of current
            iteration
        verbosity: int
            The verbosity level, 0 = silent

        Returns
        -------
        convergence: bool
            Convergence flag, true if converged

        """
        if prev_results is None:
            return False

        if verbosity > 0:
            print(f"\n{self.name}: Convergence check")
            L = max([len(v) for v in self.limits.keys()])

        ok = True
        n_states = prev_results.sizes[FC.STATE]
        self._deltas = {}
        self._conv_states = np.ones(n_states, dtype=bool)
        for v, lim in self.limits.items():
            x0 = prev_results[v].to_numpy()
            x = results[v].to_numpy()
            if v in self.wd_vars:
                a = np.abs(delta_wd(x0, x))
            else:
                a = np.abs(x - x0)
            self._deltas[v] = np.max(a)
            check = self._deltas[v]
            self._conv_states = self._conv_states & np.all(a <= lim, axis=1)
            ok = ok and (check <= lim)

            if verbosity > 0:
                r = "FAILED" if check > lim else "OK"
                print(f"  {v:<{L}}: delta = {check:.3e}, lim = {lim:.3e}  --  {r}")

        if verbosity > 0:
            print(f"Converged states: {self._conv_states.sum()}/{n_states}")

        if ok:
            self._conv_states = None

        return ok


class DefaultConv(ConvVarDelta):
    """
    Default convergence criteria.

    :group: algorithms.iterative.models

    """

    def __init__(self):
        """
        Constructor.
        """
        super().__init__(
            {
                FV.REWS: 1e-6,
                FV.TI: 1e-7,
                FV.CT: 1e-7,
            }
        )



"""
Geometries in two dimensions.
"""

from .area_geometry import AreaGeometry as AreaGeometry
from .area_geometry import InvertedAreaGeometry as InvertedAreaGeometry
from .area_geometry import AreaUnion as AreaUnion
from .area_geometry import AreaIntersection as AreaIntersection
from .polygon import ClosedPolygon as ClosedPolygon
from .circle import Circle as Circle
from .half_plane import HalfPlane as HalfPlane



import numpy as np
from scipy.interpolate import interpn

from foxes.core import States
from foxes.utils import uv2wd
from foxes.models.wake_frames.timelines import Timelines
from foxes.config import config
import foxes.variables as FV
import foxes.constants as FC


class OnePointFlowStates(States):
    """
    Time-evolving states based on horizontally
    homogeneous timeseries data

    Attributes
    ----------
    ref_xy: list of float
        The [x, y] or [x, y, z] coordinates of the base states.
        If [x, y, z] then z will serve as height
    tl_heights: list of float
        The heights at which timelines will be calculated
    dt_min: float
        The delta t value in minutes,
        if not from timeseries data
    intp_pars: dict
        Parameters for height interpolation with
        scipy.interpolate.interpn

    :group: input.states

    """

    def __init__(
        self,
        ref_xy,
        *base_states_args,
        base_states=None,
        tl_heights=None,
        dt_min=None,
        **base_states_kwargs,
    ):
        """
        Constructor.

        Parameters
        ----------
        ref_xy: list of float
            The [x, y] or [x, y, z] coordinates of the base states.
            If [x, y, z] then z will serve as height
        base_states_args: tuple, optional
            Arguments for creating the base states from
            States.new(), if not given as base_states
        base_states: foxes.core.States, optional
            The base states, representing horizontally
            homogeneous inflow
        tl_heights: list of float, optional
            The heights at which timelines will be calculated
        dt_min: float, optional
            The delta t value in minutes,
            if not from timeseries data
        base_states_kwargs: dict, optional
            Arguments for creating the base states from
            States.new(), if not given as base_states

        """
        super().__init__()
        self.ref_xy = np.array(ref_xy, dtype=config.dtype_double)
        self.heights = tl_heights
        self.base_states = base_states
        self.dt_min = dt_min

        self.intp_pars = {"fill_value": None}
        if "bounds_error" in base_states_kwargs:
            self.intp_pars["bounds_error"] = base_states_kwargs["bounds_error"]

        if base_states is not None and len(base_states_kwargs):
            raise KeyError(
                f"Base states of type '{type(base_states).__name__}' were given, cannot handle base_state_pars {list(base_states_kwargs.keys())}"
            )
        elif base_states is not None and len(base_states_args):
            raise KeyError(
                f"Base states of type '{type(base_states).__name__}' were given, cannot handle base_states_args of types {[type(a).__name__ for a in base_states_args]}"
            )
        elif base_states is None:
            self.base_states = States.new(*base_states_args, **base_states_kwargs)

    def __repr__(self):
        return f"{type(self).__name__}(base={type(self.base_states).__name__}, heights={self.heights}, dt_min={self.dt_min})"

    def sub_models(self):
        """
        List of all sub-models

        Returns
        -------
        smdls: list of foxes.core.Model
            All sub models

        """
        return [self.base_states]

    def load_data(self, algo, verbosity=0):
        """
        Load and/or create all model data that is subject to chunking.

        Such data should not be stored under self, for memory reasons. The
        data returned here will automatically be chunked and then provided
        as part of the mdata object during calculations.

        Parameters
        ----------
        algo: foxes.core.Algorithm
            The calculation algorithm
        verbosity: int
            The verbosity level, 0 = silent

        Returns
        -------
        idata: dict
            The dict has exactly two entries: `data_vars`,
            a dict with entries `name_str -> (dim_tuple, data_ndarray)`;
            and `coords`, a dict with entries `dim_name_str -> dim_array`

        """

        # find heights:
        if self.heights is None:
            if hasattr(self.base_states, "heights"):
                self.heights = self.base_states.heights
            elif len(self.ref_xy) > 2:
                self.heights = [self.ref_xy[2]]
            else:
                raise KeyError(
                    f"Cannot find 'heights' in base states of type '{type(self.base_states).__name__}', missing either `ref_xy` of type [x, y, z], or explicit value list via parameter 'tl_heights'"
                )

        # pre-calc data:
        self.WEIGHT = self.var(FV.WEIGHT)
        idata = super().load_data(algo, verbosity)
        idata["data_vars"][self.WEIGHT] = Timelines._precalc_data(
            self, algo, self.base_states, self.heights, verbosity, needs_res=True
        )

        return idata

    def size(self):
        """
        The total number of states.

        Returns
        -------
        int:
            The total number of states

        """
        return self.base_states.size()

    def index(self):
        """
        The index list

        Returns
        -------
        indices: array_like
            The index labels of states, or None for default integers

        """
        return self.base_states.index()

    def output_point_vars(self, algo):
        """
        The variables which are being modified by the model.

        Parameters
        ----------
        algo: foxes.core.Algorithm
            The calculation algorithm

        Returns
        -------
        output_vars: list of str
            The output variable names

        """
        return self.base_states.output_point_vars(algo)

    def set_running(
        self,
        algo,
        data_stash,
        sel=None,
        isel=None,
        verbosity=0,
    ):
        """
        Sets this model status to running, and moves
        all large data to stash.

        The stashed data will be returned by the
        unset_running() function after running calculations.

        Parameters
        ----------
        algo: foxes.core.Algorithm
            The calculation algorithm
        data_stash: dict, optional
            Large data stash, this function adds data here, if given.
            Key: model name. Value: dict, large model data
        sel: dict, optional
            The subset selection dictionary
        isel: dict, optional
            The index subset selection dictionary
        verbosity: int
            The verbosity level, 0 = silent

        """
        super().set_running(algo, data_stash, sel, isel, verbosity)

        if data_stash is not None and (sel is not None or isel is not None):
            data_stash[self.name]["data"] = self.timelines_data

            if isel is not None:
                self.timelines_data = self.timelines_data.isel(isel)
            if sel is not None:
                self.timelines_data = self.timelines_data.sel(sel)

    def unset_running(
        self,
        algo,
        data_stash,
        sel=None,
        isel=None,
        verbosity=0,
    ):
        """
        Sets this model status to not running, recovering large data
        from stash

        Parameters
        ----------
        algo: foxes.core.Algorithm
            The calculation algorithm
        data_stash: dict, optional
            Reconstruct model data from this stash, if given.
            Key: model name. Value: dict, large model data
        sel: dict, optional
            The subset selection dictionary
        isel: dict, optional
            The index subset selection dictionary
        verbosity: int
            The verbosity level, 0 = silent

        """
        super().unset_running(algo, data_stash, sel, isel, verbosity)

        if data_stash is not None:
            data = data_stash[self.name]
            if "data" in data:
                self.timelines_data = data.pop("data")

    def calc_states_indices(self, algo, mdata, points, hi, ref_xy):
        n_states, n_points = points.shape[:2]
        dxy = self.timelines_data["dxy"].to_numpy()[hi]

        i0 = mdata.states_i0(counter=True)
        trace_p = points[:, :, :2] - ref_xy[:, :, :2]
        trace_si = np.zeros((n_states, n_points), dtype=config.dtype_int)
        trace_si[:] = i0 + np.arange(n_states)[:, None]
        coeffs = np.full((n_states, n_points), np.nan, dtype=config.dtype_double)

        # flake8: noqa: F821
        def _eval_trace(sel, hdxy=None, hdxy0=None, trs=None):
            """Helper function that updates trace_done"""
            nonlocal coeffs

            # project onto local x direction:
            hdxy0 = dxy[trace_si[sel]] if hdxy0 is None else hdxy0
            nx = hdxy0 / np.linalg.norm(hdxy0, axis=-1)[..., None]
            projx = np.einsum("...d,...d->...", trace_p[sel], nx)

            # check for local points:
            if hdxy is None:
                seld = (projx >= 1e-10) & (projx <= 1e-10)
                if np.any(seld):
                    coeffs[sel] = np.where(seld, -1, coeffs[sel])

            # check for vicinity to reference plane:
            else:
                lx = np.einsum("...d,...d->...", hdxy, nx)
                seld = ((lx < 0) & (projx >= lx) & (projx <= 0)) | (
                    (lx > 0) & (projx >= 0) & (projx <= lx)
                )
                if np.any(seld):
                    w = projx / np.abs(lx)
                    coeffs[sel] = np.where(seld, w, coeffs[sel])

        # step backwards in time, until projection onto axis is negative:
        _eval_trace(sel=np.s_[:])
        sel = np.isnan(coeffs)
        tshift = 0
        while np.any(sel):
            trs = trace_si[sel]
            hdxy = -dxy[trs]
            trace_p[sel] += hdxy

            _eval_trace(sel, hdxy=hdxy, hdxy0=dxy[trs - tshift])

            tshift -= 1
            sel0 = np.isnan(coeffs)
            trace_si[sel0 & sel] -= 1
            sel = sel0 & (trace_si >= 0)

            del trs, sel0, hdxy

        # step forwards in time, until projection onto axis is positive:
        sel = np.isnan(coeffs)
        if np.any(sel):
            trace_p = np.where(
                sel[:, :, None], points[:, :, :2] - ref_xy[:, :, :2], trace_p
            )
            trace_si = np.where(sel, i0 + np.arange(n_states)[:, None] + 1, trace_si)

            sel &= trace_si < algo.n_states
            tshift = 1
            while np.any(sel):
                trs = trace_si[sel]
                hdxy = dxy[trs]
                trace_p[sel] += hdxy

                _eval_trace(sel, hdxy=hdxy, hdxy0=dxy[trs - tshift])

                tshift += 1
                sel0 = np.isnan(coeffs)
                trace_si[sel0 & sel] += 1
                sel = sel0 & (trace_si < algo.n_states)

                del trs, sel0, hdxy

        return trace_si, coeffs

    def calculate(self, algo, mdata, fdata, tdata):
        """
        The main model calculation.

        This function is executed on a single chunk of data,
        all computations should be based on numpy arrays.

        Parameters
        ----------
        algo: foxes.core.Algorithm
            The calculation algorithm
        mdata: foxes.core.MData
            The model data
        fdata: foxes.core.FData
            The farm data
        tdata: foxes.core.TData
            The target point data

        Returns
        -------
        results: dict
            The resulting data, keys: output variable str.
            Values: numpy.ndarray with shape
            (n_states, n_targets, n_tpoints)

        """
        # prepare:
        self.ensure_output_vars(algo, tdata)
        targets = tdata[FC.TARGETS]
        n_states, n_targets, n_tpoints = targets.shape[:3]
        n_points = n_targets * n_tpoints
        points = targets.reshape(n_states, n_points, 3)
        heights = self.timelines_data["height"].to_numpy()
        n_heights = len(heights)

        # compute states indices for all requested points:
        trace_si = []
        coeffs = []
        for hi in range(n_heights):
            s, c = self.calc_states_indices(
                algo, mdata, points, hi, self.ref_xy[None, None, :]
            )
            trace_si.append(s)
            coeffs.append(c)
            del s, c

        # flake8: noqa: F821
        def _interp_time(hi, v):
            """Helper function for interpolation bewteen states"""

            sts = trace_si[hi]
            cfs = coeffs[hi]
            data = self.timelines_data[v].to_numpy()[hi]
            out = np.zeros(sts.shape, dtype=config.dtype_double)

            sel_low = sts < 0
            if np.any(sel_low):
                out[sel_low] = data[0]

            sel_hi = sts >= algo.n_states
            if np.any(sel_hi):
                out[sel_hi] = data[algo.n_states - 1]

            sel = (~sel_low) & (~sel_hi) & (cfs <= 0)
            if np.any(sel):
                s = sts[sel]
                c = -cfs[sel]
                out[sel] = c * data[s] + (1 - c) * data[s - 1]

            sel = (~sel_low) & (~sel_hi) & (cfs > 0)
            if np.any(sel):
                s = sts[sel]
                c = cfs[sel]
                out[sel] = c * data[s - 1] + (1 - c) * data[s]

            return out

        # interpolate to heights:
        if n_heights > 1:
            ar_states = np.arange(n_states)
            ar_points = np.arange(n_points)

            crds = (heights, ar_states, ar_points)

            data = {
                v: np.stack([_interp_time(hi, v) for hi in range(n_heights)], axis=0)
                for v in self.timelines_data.data_vars.keys()
                if v != "dxy"
            }
            vres = list(data.keys())
            data = np.stack(list(data.values()), axis=-1)

            eval = np.zeros((n_states, n_points, 3), dtype=config.dtype_double)
            eval[:, :, 0] = points[:, :, 2]
            eval[:, :, 1] = ar_states[:, None]
            eval[:, :, 2] = ar_points[None, :]

            try:
                ires = interpn(crds, data, eval, **self.intp_pars)
            except ValueError as e:
                print(f"\nStates '{self.name}': Interpolation error")
                print("INPUT VARS: (heights, states, points)")
                print(
                    "DATA BOUNDS:",
                    [float(np.min(d)) for d in crds],
                    [float(np.max(d)) for d in crds],
                )
                print(
                    "EVAL BOUNDS:",
                    [float(np.min(p)) for p in eval.T],
                    [float(np.max(p)) for p in eval.T],
                )
                print(
                    "\nMaybe you want to try the option 'bounds_error=False'? This will extrapolate the data.\n"
                )
                raise e
            del crds, eval, data, ar_points, ar_states

            results = {}
            for v in self.output_point_vars(algo):
                if v not in [FV.WS, FV.WD]:
                    results[v] = ires[:, :, vres.index(v)]
                elif v not in results:
                    uv = np.stack(
                        [ires[:, :, vres.index("U")], ires[:, :, vres.index("V")]],
                        axis=-1,
                    )
                    results = {FV.WD: uv2wd(uv), FV.WS: np.linalg.norm(uv, axis=-1)}
                    del uv

        # no dependence on height:
        else:
            results = {}
            for v in self.output_point_vars(algo):
                if v not in [FV.WS, FV.WD]:
                    results[v] = _interp_time(hi, v)
                elif v not in results:
                    uv = np.stack(
                        [_interp_time(hi, "U"), _interp_time(hi, "V")], axis=-1
                    )
                    results = {FV.WD: uv2wd(uv), FV.WS: np.linalg.norm(uv, axis=-1)}
                    del uv

        # set weights:
        tdata[FV.WEIGHT] = mdata[self.WEIGHT][:, None, None]
        tdata.dims[FV.WEIGHT] = (FC.STATE, FC.TARGET, FC.TPOINT)

        return {
            v: d.reshape(n_states, n_targets, n_tpoints) for v, d in results.items()
        }


class OnePointFlowTimeseries(OnePointFlowStates):
    """
    Inhomogeneous inflow from homogeneous timeseries data
    at one point

    :group: input.states

    """

    def __init__(self, ref_xy, *args, tl_heights=None, **kwargs):
        """
        Constructor.

        Parameters
        ----------
        ref_xy: list of float
            The [x, y] or [x, y, z] coordinates of the base states.
            If [x, y, z] then z will serve as height
        args: tuple, optional
            Parameters for the base class
        tl_heights: list of float, optional
            The heights at which timelines will be calculated
        kwargs: dict, optional
            Parameters for the base class

        """
        if tl_heights is None and len(ref_xy) < 3:
            tl_heights = [100.0]
        super().__init__(
            ref_xy,
            *args,
            tl_heights=tl_heights,
            states_type="Timeseries",
            **kwargs,
        )


class OnePointFlowMultiHeightTimeseries(OnePointFlowStates):
    """
    Inhomogeneous inflow from height dependent homogeneous
    timeseries data at one point

    :group: input.states

    """

    def __init__(self, *args, **kwargs):
        """
        Constructor.

        Parameters
        ----------
        args: tuple, optional
            Parameters for the base class
        kwargs: dict, optional
            Parameters for the base class

        """
        super().__init__(*args, states_type="MultiHeightTimeseries", **kwargs)


class OnePointFlowMultiHeightNCTimeseries(OnePointFlowStates):
    """
    Inhomogeneous inflow from height dependent homogeneous
    timeseries data at one point based on NetCDF input

    :group: input.states

    """

    def __init__(self, *args, **kwargs):
        """
        Constructor.

        Parameters
        ----------
        args: tuple, optional
            Parameters for the base class
        kwargs: dict, optional
            Parameters for the base class

        """
        super().__init__(*args, states_type="MultiHeightNCTimeseries", **kwargs)



"""
Wake models.
"""

from .dist_sliced import DistSlicedWakeModel as DistSlicedWakeModel
from .axisymmetric import AxisymmetricWakeModel as AxisymmetricWakeModel
from .top_hat import TopHatWakeModel as TopHatWakeModel
from .gaussian import GaussianWakeModel as GaussianWakeModel

from . import wind as wind
from . import ti as ti
from . import induction as induction



import numpy as np
import pandas as pd

from foxes.core import Turbine
from foxes.config import get_input_path
from foxes.models.turbine_types import PCtFile


def add_from_eww(
    farm,
    data_source,
    filter={},
    mbook=None,
    csv_dir=None,
    rho=1.225,
    verbosity=1,
    pct_pars={},
    **turbine_parameters,
):
    """
    Add turbines to wind farm via EuroWindWakes database csv input file.

    Parameters
    ----------
    farm: foxes.WindFarm
        The wind farm
    data_source: str or pandas.DataFrame
        The input csv file or data source
    filter: list of tuple, optional
        A list of filters to apply to the dataframe,
        e.g. ("wind_farm": ["Farm1", "Farm2"]),
        or ("latitude", ">=54.1"). For range filtering, use strings
        that start with ">=", "<=", ">", "<". For exact matches, use single values.
        For multiple matches, use lists, tuples or sets.
    mbook: foxes.ModelBook, optional
        The model book, only needed if csv_dir is specified
    csv_dir: str, optional
        The csv file directory, containing turbine type data files
    rho: float
        The air density for the turbine types, if csv_dir is given
    verbosity: int
        The verbosity level, 0 = silent
    pct_pars: dict
        Additional parameters for the PCtFile constructor
    turbine_parameters: dict, optional
        Additional parameters are forwarded to the WindFarm.add_turbine().

    Examples
    --------
    Data source format:

    >>>    ,wind_farm,oem_manufacturer,latitude,longitude,country,rated_power,rotor_diameter,hub_height,turbine_type,commissioning_date
    >>>    0,Aberdeen Offshore Wind Farm,Vestas,57.230095,-1.9742404,United Kingdom,8.4,164.0,108.5,V164-8.4 MW,2018-09
    >>>    1,Aberdeen Offshore Wind Farm,Vestas,57.2235827,-2.0127432,United Kingdom,8.4,164.0,108.5,V164-8.4 MW,2018-09
    >>>    2,Aberdeen Offshore Wind Farm,Vestas,57.2169301,-2.0055697,United Kingdom,8.4,164.0,108.5,V164-8.4 MW,2018-09
    >>>    ...

    :group: input.farm_layout

    """
    assert farm.data_is_lonlat, "Require `input_is_lonlat=True` in WindFarm constructor"

    if isinstance(data_source, pd.DataFrame):
        data = data_source
    else:
        if verbosity:
            print("Reading file", data_source)
        pth = get_input_path(data_source)
        data = pd.read_csv(pth, index_col=0, parse_dates=["commissioning_date"])

    comm_dates = data["commissioning_date"].str.strip()
    comm_dates = pd.to_datetime(comm_dates)
    data["comm_year"] = comm_dates.dt.year
    data["comm_month"] = comm_dates.dt.month
    data["comm_day"] = comm_dates.dt.day

    ttypes = data["turbine_type"].unique()
    if csv_dir is not None:
        if mbook is None:
            raise ValueError("Model book must be given if csv_dir is specified")
        csv_dir = get_input_path(csv_dir)
        csv_files = sorted(list(csv_dir.glob("*.csv")))
        csv_names = [f.stem for f in csv_files]
        if verbosity > 0:
            print(f"Reading {len(csv_files)} CSV files from {csv_dir}")
        csv_map = []
        ntt = []
        for t in ttypes:
            if t not in csv_names:
                t = t.replace(" ", "_").replace("/", "_")
            if t not in csv_names:
                raise ValueError(f"Turbine type {t} not found in {csv_dir}")
            ntt.append(t)
            csv_map.append(csv_files[csv_names.index(t)])
        ttypes = ntt
        for t in ttypes:
            mbook.turbine_types[t] = None

    else:
        if verbosity > 0:
            print(
                "No csv_dir specified, assuming turbine types correspond to model book names"
            )

    if filter is not None:
        assert isinstance(filter, list), (
            f"filter must be a list, got {type(filter).__name__}"
        )
        for fi, fdata in enumerate(filter):
            assert isinstance(fdata, tuple), (
                f"filter items must be tuples, got {type(fdata).__name__} at position {fi}"
            )
            k, val = fdata
            assert k in data.columns, f"Column {k} not in data, found {data.columns}"
            if isinstance(val, str):
                if val.startswith(">="):
                    data = data[data[k] >= float(val[2:])]
                elif val.startswith("<="):
                    data = data[data[k] <= float(val[2:])]
                elif val.startswith(">"):
                    data = data[data[k] > float(val[1:])]
                elif val.startswith("<"):
                    data = data[data[k] < float(val[1:])]
                if verbosity > 0:
                    print(f"Applying filter {k}{val}, now {len(data.index)} turbines")
            elif isinstance(val, (list, tuple, set)):
                data = data[data[k].isin(val)]
                if verbosity > 0:
                    print(
                        f"Applying filter {k} in {val}, now {len(data.index)} turbines"
                    )
            else:
                data = data[data[k] == val]
                if verbosity > 0:
                    print(f"Applying filter {k}=={val}, now {len(data.index)} turbines")

    if verbosity > 0:
        print(f"Selected {len(data.index)} turbines from the following wind farms:")
        for wf, g in data.groupby("wind_farm"):
            print(f"  {wf} ({g['commissioning_date'].min()}): {len(g.index)} turbines")

    farms = []
    tmodels = turbine_parameters.pop("turbine_models", [])
    for i0, i in enumerate(data.index):
        fname = str(data.loc[i, "wind_farm"]).replace(" ", "_")
        if fname not in farms:
            farms.append(fname)
            j = 0

        ttype = data.loc[i, "turbine_type"]
        if ttype not in ttypes:
            ttype = ttype.replace(" ", "_").replace("/", "_")
        if csv_dir is not None and mbook.turbine_types[ttype] is None:
            j = ttypes.index(ttype)
            if verbosity > 0:
                print(f"Creating turbine type: {ttype} from file {csv_map[j].name}")
            pars = dict(col_P="power")
            pars.update(pct_pars)
            mbook.turbine_types[ttype] = PCtFile(csv_map[j], rho=rho, **pars)

        lonlat = data.loc[i, ["longitude", "latitude"]].to_numpy(np.float64)
        farm.add_turbine(
            Turbine(
                name=f"{fname}_{j}",
                index=i0,
                xy=lonlat,
                H=data.loc[i, "hub_height"],
                D=data.loc[i, "rotor_diameter"],
                turbine_models=[ttype] + tmodels,
                wind_farm_name=fname,
                **turbine_parameters,
            ),
            verbosity=verbosity,
        )

        j += 1

    farm.lock(verbosity=verbosity)



import numpy as np

from foxes.core import FarmDataModel, TData
import foxes.variables as FV
import foxes.constants as FC
from foxes.config import config


class InitFarmData(FarmDataModel):
    """
    Sets basic turbine data and applies downwind order

    :group: algorithms.downwind.models

    """

    def __init__(self):
        """
        Constructor.
        """
        super().__init__(pre_rotor=True)

    def output_farm_vars(self, algo):
        """
        The variables which are being modified by the model.

        Parameters
        ----------
        algo: foxes.core.Algorithm
            The calculation algorithm

        Returns
        -------
        output_vars: list of str
            The output variable names

        """
        return [
            FV.X,
            FV.Y,
            FV.H,
            FV.D,
            FV.WD,
            FV.YAW,
            FV.ORDER,
            FV.ORDER_SSEL,
            FV.ORDER_INV,
        ]

    def calculate(self, algo, mdata, fdata):
        """
        The main model calculation.

        This function is executed on a single chunk of data,
        all computations should be based on numpy arrays.

        Parameters
        ----------
        algo: foxes.core.Algorithm
            The calculation algorithm
        mdata: foxes.core.Data
            The model data
        fdata: foxes.core.Data
            The farm data

        Returns
        -------
        results: dict
            The resulting data, keys: output variable str.
            Values: numpy.ndarray with shape (n_states, n_turbines)

        """
        # prepare:
        n_states = fdata.n_states
        n_turbines = algo.n_turbines

        # add and set X, Y, H, D:
        fdata.add(
            FV.TXYH,
            np.zeros((n_states, n_turbines, 3), dtype=config.dtype_double),
            (FC.STATE, FC.TURBINE, FC.XYH),
        )
        fdata.add(
            FV.D,
            np.zeros((n_states, n_turbines), dtype=config.dtype_double),
            (FC.STATE, FC.TURBINE),
        )
        for ti, t in enumerate(algo.farm.turbines):
            if len(t.xy.shape) == 1:
                fdata[FV.TXYH][:, ti, :2] = t.xy[None, :]
            else:
                i0 = fdata.states_i0(counter=True)
                s = np.s_[i0 : i0 + fdata.n_states]
                fdata[FV.TXYH][:, ti, :2] = t.xy[s]

            H = t.H
            if H is None:
                H = algo.farm_controller.turbine_types[ti].H
            fdata[FV.TXYH][:, ti, 2] = H

            D = t.D
            if D is None:
                D = algo.farm_controller.turbine_types[ti].D
            fdata[FV.D][:, ti] = D

        # calc WD at rotor centres:
        svrs = algo.states.output_point_vars(algo)
        tdata = TData.from_points(points=fdata[FV.TXYH], variables=svrs)
        sres = algo.states.calculate(algo, mdata, fdata, tdata)
        fdata.add(
            FV.WD,
            sres[FV.WD][:, :, 0],
            (FC.STATE, FC.TURBINE),
        )
        fdata.add(
            FV.AMB_WD,
            fdata[FV.WD].copy(),
            (FC.STATE, FC.TURBINE),
        )
        del tdata, sres, svrs

        # calculate downwind order:
        order = algo.wake_frame.calc_order(algo, mdata, fdata)
        ssel = np.zeros_like(order)
        ssel[:] = np.arange(n_states)[:, None]

        # apply downwind order to all data:
        for data in [fdata, mdata]:
            for k in data.keys():
                if (
                    k not in [FV.X, FV.Y, FV.H]
                    and tuple(data.dims[k][:2]) == (FC.STATE, FC.TURBINE)
                    and np.any(data[k] != data[k][0, 0, None, None])
                ):
                    data[k][:] = data[k][ssel, order]

        # add derived data:
        for i, v in enumerate([FV.X, FV.Y, FV.H]):
            fdata.add(
                v,
                fdata[FV.TXYH][:, :, i],
                (FC.STATE, FC.TURBINE),
            )
        fdata.add(
            FV.YAW,
            fdata[FV.WD].copy(),
            (FC.STATE, FC.TURBINE),
        )
        fdata.add(
            FV.ORDER,
            order,
            (FC.STATE, FC.TURBINE),
        )
        fdata.add(
            FV.ORDER_SSEL,
            ssel,
            (FC.STATE, FC.TURBINE),
        )
        fdata.add(
            FV.ORDER_INV,
            np.zeros_like(order),
            (FC.STATE, FC.TURBINE),
        )
        fdata[FV.ORDER_INV][ssel, order] = np.arange(n_turbines)[None, :]

        return {v: fdata[v] for v in self.output_farm_vars(algo)}



import argparse
from pathlib import Path

from foxes.core import Algorithm
from foxes.models import ModelBook
from foxes.utils import import_module, Dict

from .read_site import read_site
from .read_farm import read_farm
from .read_attributes import read_attributes
from .read_outputs import read_outputs
from ..dict import run_dict


def windio_file2dict(yml_file, verbosity=1):
    """
    Read windio yaml file and translate to foxes input data dictionary

    Parameters
    ----------
    yml_file: pathlib.Path or str
        The windio yaml file
    verbosity: int
        The verbosity level, 0 = silent

    Returns
    -------
    wio_dict: foxes.utils.Dict
        The windio data dictionary

    :group: input.yaml.windio

    """

    wio_file = Path(yml_file)
    if verbosity > 0:
        print(f"Reading windio file {wio_file}")

    windio = import_module(
        "windIO",
        pip_hint="pip install git+https://github.com/EUFLOW/windIO@master",
        conda_hint="",
    )

    return Dict(windio.load_yaml(wio_file), _name="windio")


def read_windio_dict(wio_dict, verbosity=1, **algo_kwargs):
    """
    Translate windio data to foxes input data

    Parameters
    ----------
    wio_dict: foxes.utils.Dict
        The windio data
    ret_algo: bool
        Whether to return the algorithm object, otherwise
        return its parameters
    verbosity: int
        The verbosity level, 0 = silent
    algo_kwargs: dict, optional
        Additional keyword arguments for the algorithm

    Returns
    -------
    idict: foxes.utils.Dict or dict
        The foxes input data dictionary
    algo: foxes.core.Algorithm
        The algorithm
    odir: pathlib.Path
        The output directory

    :group: input.yaml.windio

    """

    def _print(*args, level=1, **kwargs):
        if verbosity >= level:
            print(*args, **kwargs)

    if not isinstance(wio_dict, Dict):
        tmp = Dict(_name="windio")
        for k, d in wio_dict.items():
            tmp[k] = d
        wio_dict = tmp

    _print("Reading windio data")
    _print("  Name:", wio_dict.pop_item("name", None))
    _print("  Contents:", [k for k in wio_dict.keys()])

    idict = Dict(
        wind_farm=Dict(_name="wio2fxs.farm"),
        algorithm=Dict(
            algo_type="Downwind",
            wake_models=[],
            farm_controller="farm_cntrl",
            _name="wio2fxs.algorithm",
            verbosity=verbosity - 3,
        ),
        calc_farm=Dict(run=True, _name="wio2fxs.calc_farm"),
        _name="wio2fxs",
    )

    mbook = ModelBook()
    states = read_site(wio_dict, verbosity)
    farm = read_farm(wio_dict, mbook, verbosity)

    wio_attrs = wio_dict["attributes"]
    read_attributes(wio_attrs, idict, mbook, verbosity=verbosity)

    # special case WeibullPointCloud:
    if (
        type(states).__name__ == "WeibullPointCloud"
        and idict["algorithm"]["rotor_model"] == "centre"
    ):
        _print(
            "Found WeibullPointCloud states, changing rotor model from 'centre' to 'direct_mdata'",
            level=3,
        )
        idict["algorithm"]["rotor_model"] = "direct_mdata"

    idict["algorithm"].update(algo_kwargs)
    if verbosity > 1:
        print("\nFinal input dictionary:\n\n", idict, "\n")
    algo = Algorithm.new(
        farm=farm, states=states, mbook=mbook, **idict.pop_item("algorithm")
    )

    odir = None
    if "model_outputs_specification" in wio_attrs:
        outputs = wio_attrs["model_outputs_specification"]
        odir = read_outputs(outputs, idict, algo, verbosity=verbosity)

    return idict, algo, odir


def read_windio_file(yml_file, ret_wio=False, verbosity=1, **algo_kwargs):
    """
    Read windio yaml file and translate to foxes input data

    Parameters
    ----------
    yml_file: pathlib.Path or str
        The windio yaml file
    ret_wio: bool
        Whether to return the windio data dictionary as well
    verbosity: int
        The verbosity level, 0 = silent
    algo_kwargs: dict, optional
        Additional keyword arguments for the algorithm

    Returns
    -------
    idict: foxes.utils.Dict or dict
        The foxes input data dictionary
    algo: foxes.core.Algorithm
        The algorithm
    odir: pathlib.Path
        The output directory

    :group: input.yaml.windio

    """
    wio = windio_file2dict(yml_file, verbosity)
    idict, algo, odir = read_windio_dict(wio, verbosity=verbosity, **algo_kwargs)

    if ret_wio:
        return idict, algo, odir, wio
    else:
        return idict, algo, odir


def foxes_windio():
    """
    Command line tool for running foxes from windio yaml file input.

    Examples
    --------
    >>> foxes_windio input.yaml

    :group: input.yaml.windio

    """

    parser = argparse.ArgumentParser()
    parser.add_argument(
        "yml_file",
        help="The windio yaml file",
    )
    parser.add_argument("-o", "--output_dir", help="The output directory", default=None)
    parser.add_argument("-r", "--rotor", help="The rotor model", default="centre")
    parser.add_argument(
        "-p", "--pwakes", help="The partial wakes models", default="centre", nargs="+"
    )
    parser.add_argument(
        "-w",
        "--wakes",
        help="The wake models",
        default=["Jensen_linear_k007"],
        nargs="+",
    )
    parser.add_argument("-f", "--frame", help="The wake frame", default="rotor_wd")
    parser.add_argument("-e", "--engine", help="The engine", default=None)
    parser.add_argument(
        "-n", "--n_procs", help="The number of processes", default=None, type=int
    )
    parser.add_argument(
        "-c",
        "--chunksize_states",
        help="The chunk size for states",
        default=None,
        type=int,
    )
    parser.add_argument(
        "-C",
        "--chunksize_points",
        help="The chunk size for points",
        default=5000,
        type=int,
    )
    parser.add_argument(
        "-it", "--iterative", help="Use iterative algorithm", action="store_true"
    )
    parser.add_argument(
        "-nf", "--nofig", help="Do not show figures", action="store_true"
    )
    parser.add_argument(
        "-v",
        "--verbosity",
        help="The verbosity level, 0 = silent",
        type=int,
        default=1,
    )
    args = parser.parse_args()

    if (
        args.engine is not None
        or args.n_procs is not None
        or args.chunksize_states is not None
        or args.chunksize_points is not None
    ):
        epars = dict(
            engine_type=args.engine,
            n_procs=args.n_procs,
            chunk_size_states=args.chunksize_states,
            chunk_size_points=args.chunksize_points,
            verbosity=args.verbosity,
        )
    else:
        epars = None

    wio_file = Path(args.yml_file)
    idict, algo, odir = read_windio_file(wio_file, verbosity=args.verbosity)

    if args.output_dir is not None:
        odir = args.odir

    run_dict(
        idict,
        algo=algo,
        rotor_model=args.rotor,
        partial_wakes=args.pwakes,
        wake_models=args.wakes,
        wake_frame=args.frame,
        engine_pars=epars,
        iterative=args.iterative,
        input_dir=wio_file.parent,
        output_dir=odir,
        verbosity=args.verbosity,
    )



"""
Wind deficit wake models.
"""

from .jensen import JensenWake as JensenWake
from .bastankhah14 import Bastankhah2014 as Bastankhah2014
from .bastankhah16 import Bastankhah2016Model as Bastankhah2016Model
from .bastankhah16 import Bastankhah2016 as Bastankhah2016
from .turbopark import TurbOParkWake as TurbOParkWake
from .turbopark import TurbOParkWakeIX as TurbOParkWakeIX
